###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file describe the virtual cell layout widget used in
Parameter Exploration Tab """
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.inspector import PipelineInspector
from vistrails.gui.common_widgets import QToolWindowInterface
from vistrails.gui.paramexplore.pe_pipeline import QAnnotatedPipelineView
from vistrails.gui.theme import CurrentTheme
import copy
import os.path

###############################################################################

def split_camel_case(text):
    if not text:
        return u''
    lines = []
    # State machine!
    #                                 empty?
    #          initial ------------------------>
    #           |    \             return empty
    #           |      \
    #           |        \
    #           |          \
    #           |            \
    #         U | mark       L \ mark
    #           |                \
    #          \/        L       \/
    #          1  ---------------> 2 ----\
    #       /  /\ |\             / |\    | L
    #     /    |   \      U     /   \___/
    #    /     |    \-----------
    #   |      |      emit, mark
    #   |      |
    # U |    L | emit -1
    #   |      | mark -1
    #   |     /
    #   |   /
    #  \/ /
    #    3 ----\
    #    |\    | U
    #     \___/
    #
    state = 0
    for i, c in enumerate(text):
        u = c.isupper()
        if state == 1:
            state = 3 if u else 2
        elif state == 2 and u:
            state = 1
            lines.append(text[mark:i])
            mark = i
        elif state == 3 and not u:
            lines.append(text[mark:i-1])
            mark = i-1
            state = 1
        elif state == 0:
            state = 1 if u else 2
            mark = i
    lines.append(text[mark:])

    return u'\n'.join(lines)

###############################################################################

def decodeConfiguration(pipeline, cells):
    """ decodeConfiguration(pipeline: Pipeline,
        cells: configuration) -> decoded cells
    Convert cells of type [{(type,id): (row, column)}) to
    (mId, row, col) in a particular pipeline

    """
    decodedCells = []
    inspector = PipelineInspector()
    inspector.inspect_spreadsheet_cells(pipeline)
    inspector.inspect_ambiguous_modules(pipeline)
    orig_pipeline = pipeline

    for id_list in inspector.spreadsheet_cells:
        pipeline = orig_pipeline
        id_iter = iter(id_list)
        m = pipeline.modules[id_iter.next()]
        m_id = m.id
        for m_id in id_iter:
            pipeline = m.pipeline
            m = pipeline.modules[m_id]
        name = m.name

        if len(id_list) == 1 and m_id in inspector.annotated_modules:
            idx = inspector.annotated_modules[m_id]
        elif tuple(id_list) in inspector.annotated_modules:
            idx = inspector.annotated_modules[tuple(id_list)]
        else:
            idx = -1
        (vRow, vCol) = cells[(name, idx)]
        if len(id_list) == 1:
            decodedCells.append((m_id, vRow, vCol))
        else:
            decodedCells.append((tuple(id_list), vRow, vCol))
    return decodedCells

def positionPipelines(sheetPrefix, sheetCount, rowCount, colCount,
                       pipelines, cells, controller):
    """ _positionPipelines(sheetPrefix: str, sheetCount: int, rowCount: int,
                           colCount: int, pipelines: list of Pipeline,
                           cells: List, controller: VistrailCintroller)
                           -> list of Pipelines
    Apply the virtual cell location to a list of pipelines in a
    parameter exploration given that pipelines has multiple chunk
    of sheetCount x rowCount x colCount cells

    """

    # at this point, we know that we have the spreadsheet loaded
    from vistrails.packages.spreadsheet.spreadsheet_execute import \
        assignPipelineCellLocations

    modifiedPipelines = []
    pipelinePositions = []
    for pId in xrange(len(pipelines)):
        root_pipeline = copy.copy(pipelines[pId])
        col = pId % colCount
        row = (pId // colCount) % rowCount
        sheet = (pId // (colCount*rowCount)) % sheetCount

        decodedCells = decodeConfiguration(root_pipeline, cells)
        vRCount = (max(c[1] for c in decodedCells) + 1) if len(decodedCells) else 1
        vCCount = (max(c[2] for c in decodedCells) + 1) if len(decodedCells) else 1
        # still need to go through each separately
        for (id_list, vRow, vCol) in decodedCells:
            sheet_name = "%s %d" % (sheetPrefix, sheet)
            min_row_count = rowCount * vRCount
            min_col_count = colCount * vCCount
            real_row = row*vRCount+vRow+1
            real_col = col*vCCount+vCol+1
            root_pipeline = \
                assignPipelineCellLocations(root_pipeline, sheet_name,
                                            real_row, real_col,
                                            [id_list], min_row_count,
                                            min_col_count)

        modifiedPipelines.append(root_pipeline)
        pipelinePositions.append((row, col, sheet))
    return modifiedPipelines, pipelinePositions

def assembleThumbnails(images, name, background='#000000'):
    """ assembleThumbnails(images {(sheet, row, col):filename}, name: 'str',
                           background: str)"""
                           
    bgcolor = QtGui.QColor()
    bgcolor.setNamedColor(background)
    maxRow = max(i[0] for i in images) + 1
    maxCol = max(i[1] for i in images) + 1
    maxSheet = max(i[2] for i in images) + 1
    
    w = h = 0
    sheets = []
    painter = QtGui.QPainter()
    
    # paint thumbnails on correct positions
    for pos, filename in images.iteritems():
        row, col, sheet = pos

        if not os.path.exists(filename + '.png'):
            continue
        pixmap = QtGui.QPixmap(filename)

        # use first pixmap to calculate spreadsheet size
        # this assumes all pixmaps have same size
        if not w:
            w, h = pixmap.width(), pixmap.height()
            sizeX = w*maxCol
            sizeY = h*maxRow
            # init sheets with correct background
            for i in xrange(maxSheet):
                _sheet = QtGui.QPixmap(sizeX, sizeY)
                sheets.append(_sheet)
                painter.begin(_sheet)
                painter.fillRect(0, 0, sizeX, sizeY, bgcolor)
                painter.end()

        sheet = sheets[sheet]
        painter.begin(sheet)
        painter.drawPixmap(col*w, h*row, w, h, pixmap.scaled(w, h,
                                transformMode=QtCore.Qt.SmoothTransformation))
        painter.end()
        # draw spreadsheet grid
        #painter.setPen(QtCore.Qt.black)
        #for x in xrange(rows+1):
        #    painter.drawLine(x*h-0.01, 0, x*h-0.01, size)
        #for y in xrange(cols+1):
        #    painter.drawLine(0, y*w-0.01, size, y*w-0.01)
        #self.setIcon(0, QtGui.QIcon(pic))
    
    for i in xrange(len(sheets)):
        sheet = sheets[i]
        filename = '%s_%s.png' % (name, i)
        sheet.save(filename)
        
class QVirtualCellWindow(QtGui.QFrame, QToolWindowInterface):
    """
    QVirtualCellWindow contains a caption, a virtual cell
    configuration
    
    """
    def __init__(self, parent=None):
        """ QVirtualCellWindow(parent: QWidget) -> QVirtualCellWindow
        Initialize the widget

        """
        QtGui.QFrame.__init__(self, parent)
        self.setFrameShape(QtGui.QFrame.StyledPanel)
        self.setWindowTitle('Spreadsheet Virtual Cell')
        vLayout = QtGui.QVBoxLayout(self)
        vLayout.setMargin(2)
        vLayout.setSpacing(0)
        self.setLayout(vLayout)
        
        label = QtGui.QLabel('Arrange the cell(s) below to construct'
                             ' a virtual cell')
        font = QtGui.QFont(label.font())
        label.setFont(font)
        label.setWordWrap(True)        
        vLayout.addWidget(label)

        hLayout = QtGui.QVBoxLayout()
        hLayout.setMargin(0)
        hLayout.setSpacing(0)
        vLayout.addLayout(hLayout)
        self.config = QVirtualCellConfiguration()
        self.config.setSizePolicy(QtGui.QSizePolicy.Maximum,
                                  QtGui.QSizePolicy.Maximum)
        hLayout.addWidget(self.config)
        hPadWidget = QtGui.QWidget()
        hLayout.addWidget(hPadWidget)
        hPadWidget.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                 QtGui.QSizePolicy.Ignored)

        vPadWidget = QtGui.QWidget()
        vLayout.addWidget(vPadWidget)
        vPadWidget.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                QtGui.QSizePolicy.Expanding)

        self.inspector = PipelineInspector()
        self.pipeline = None

    def updateVirtualCell(self, pipeline):        
        """ updateVirtualCell(pipeline: QPipeline) -> None
        Setup the virtual cells given a pipeline
        
        """
        self.pipeline = pipeline
        if self.pipeline and self.pipeline.is_valid:
            self.inspector.inspect_spreadsheet_cells(self.pipeline)
            self.inspector.inspect_ambiguous_modules(self.pipeline)
            cells = []
            for id_list in self.inspector.spreadsheet_cells:
                pipeline = self.pipeline
                id_iter = iter(id_list)
                m = pipeline.modules[id_iter.next()]
                m_id = m.id
                for m_id in id_iter:
                    pipeline = m.pipeline
                    m = pipeline.modules[m_id]
                    
                name = m.name
                if len(id_list) == 1 and \
                        m_id in self.inspector.annotated_modules:
                    cells.append((name, self.inspector.annotated_modules[m_id]))
                elif tuple(id_list) in self.inspector.annotated_modules:
                    cells.append((name, self.inspector.annotated_modules[ \
                                tuple(id_list)]))
                else:
                    cells.append((name, -1))
            self.config.configVirtualCells(cells)
        else:

            self.config.clear()

    def getConfiguration(self):
        """ getConfiguration() -> info (see below)
        Return the current configuration of the virtual cell. The
        information is:
        info = (rowCount, columnCount,
                {(type,id): (row, column)})
        """
        return self.config.getConfiguration()
                
    def setConfiguration(self, info):
        """ setConfiguration(info) -> None (see below)
        Set the configuration of the virtual cell. The
        information is:
        info = {(type, id): (row, column)}
          or
        info = (rowCount, columnCount,
                {(type, id): (row, column)})
        The second form is allowed so that the output of
        getConfiguration could be passed directly to
        setConfiguration (the dimensions aren't used).
        """
        self.config.setConfiguration(info)

class QVirtualCellConfiguration(QtGui.QWidget):
    """
    QVirtualCellConfiguration is a widget provide a virtual layout of
    the spreadsheet cell. Given a number of cells want to layout, it
    will let users interactively select where to put a cell in a table
    layout to construct a virtual cell out of that.
    
    """
    def __init__(self, parent=None):
        """ QVirtualCellConfiguration(parent: QWidget)
                                      -> QVirtualCellConfiguration
        Initialize the widget

        """
        QtGui.QWidget.__init__(self, parent)
        self.rowCount = 1
        self.colCount = 1
        gridLayout = QtGui.QGridLayout(self)
        gridLayout.setSpacing(0)
        self.setLayout(gridLayout)
        label = QVirtualCellLabel('')
        self.layout().addWidget(label, 0, 0, 1, 1, QtCore.Qt.AlignCenter)
        self.cells = [[label]]
        self.numCell = 1

    def clear(self):
        """ clear() -> None
        Remove and delete all widgets in self.gridLayout
        
        """
        while True:
            item = self.layout().takeAt(0)
            if item is None:
                break
            self.disconnect(item.widget(),
                            QtCore.SIGNAL('finishedDragAndDrop'),
                            self.compressCells)
            item.widget().deleteLater()
            del item
        self.cells = []
        self.numCell = 0

    def configVirtualCells(self, cells):
        """ configVirtualCells(cells: [(str, int)]) -> None        
        Given a list of cell types and ids, this will clear old
        configuration and start a fresh one.
        
        """
        self.clear()
        self.numCell = len(cells)
        row = []
        for i in xrange(self.numCell):
            label = QVirtualCellLabel(*cells[i])
            row.append(label)
            self.layout().addWidget(label, 0, i, 1, 1, QtCore.Qt.AlignCenter)
            self.connect(label, QtCore.SIGNAL('finishedDragAndDrop'),
                         self.compressCells)
        self.cells.append(row)

        for r in xrange(self.numCell-1):
            row = []
            for c in xrange(self.numCell):
                label = QVirtualCellLabel()
                row.append(label)
                self.layout().addWidget(label, r+1, c, 1, 1,
                                        QtCore.Qt.AlignCenter)
                self.connect(label, QtCore.SIGNAL('finishedDragAndDrop'),
                             self.compressCells)
            self.cells.append(row)

    def compressCells(self):
        """ compressCells() -> None
        Eliminate all blank cells
        
        """
        # Check row by row first
        visibleRows = []
        for r in xrange(self.numCell):
            row = self.cells[r]
            hasRealCell = [True for label in row if label.type]!=[]
            if hasRealCell:                
                visibleRows.append(r)

        # Move rows up
        for i in xrange(len(visibleRows)):
            for c in xrange(self.numCell):
                label = self.cells[visibleRows[i]][c]
                if label.type is None:
                    label.type = ''
                self.cells[i][c].setCellData(label.type, label.id)

        # Now check column by column        
        visibleCols = []
        for c in xrange(self.numCell):
            hasRealCell = [True
                           for r in xrange(self.numCell)
                           if self.cells[r][c].type]!=[]
            if hasRealCell:
                visibleCols.append(c)
                    
        # Move columns left
        for i in xrange(len(visibleCols)):
            for r in xrange(self.numCell):
                label = self.cells[r][visibleCols[i]]
                if label.type is None:
                    label.type = ''
                self.cells[r][i].setCellData(label.type, label.id)

        # Clear redundant rows
        for i in xrange(self.numCell-len(visibleRows)):
            for label in self.cells[i+len(visibleRows)]:
                label.setCellData(None, -1)
                
        # Clear redundant columns
        for i in xrange(self.numCell-len(visibleCols)):
            for r in xrange(self.numCell):
                self.cells[r][i+len(visibleCols)].setCellData(None, -1)

    def getConfiguration(self):
        """ getConfiguration() -> info (see below)
        Return the current configuration of the virtual cell. The
        information is:
        info = (rowCount, columnCount,
                {(type, id): (row, column)})
        """
        result = {}
        rCount = 0
        cCount = 0
        for r in xrange(self.numCell):
            for c in xrange(self.numCell):
                cell = self.cells[r][c]
                if cell.type:
                    result[(cell.type, cell.id)] = (r, c)
                    if r+1>rCount: rCount = r+1
                    if c+1>cCount: cCount = c+1
        return (rCount, cCount, result)

    def setConfiguration(self, info):
        """ setConfiguration(info) -> None (see below)
        Set the configuration of the virtual cell. The
        information is:
        info = {(type, id): (row, column)}
          or
        info = (rowCount, columnCount,
                {(type, id): (row, column)})
        The second form is allowed so that the output of
        getConfiguration could be passed directly to
        setConfiguration (the dimensions aren't used).
        """
        if isinstance(info, dict):
            result = info
        else:
            rCount, cCount, result = info
        # Reset the layout of the virtual cell to default state
        config_cells = []
        for cell_type, cell_id in result.iterkeys():
            config_cells.append((cell_type, cell_id))
        self.configVirtualCells(config_cells)
        # Unset the 0th row types/ids, since they're auto-set for the default state
        for c in xrange(len(config_cells)):
            self.cells[0][c].setCellData('', -1)
        # Set the new types/ids
        for cell_type, cell_id in config_cells:
            row, col = result[(cell_type, cell_id)]
            self.cells[row][col].setCellData(cell_type, cell_id)
        # Compress to properly reset newly empty cells
        self.compressCells()

class QVirtualCellLabel(QtGui.QLabel):
    """
    QVirtualCellLabel is a label represent a cell inside a cell. It
    has rounded shape with a caption text
    
    """
    def __init__(self, label=None, id=-1, parent=None):
        """ QVirtualCellLabel(text: QString, id: int,
                              parent: QWidget)
                              -> QVirtualCellLabel
        Construct the label image

        """
        QtGui.QLabel.__init__(self, parent)
        self.setMargin(2)
        self.cellType = None
        self.setCellData(label, id)
        self.setAcceptDrops(True)
        self.setFrameStyle(QtGui.QFrame.Panel)
        self.palette().setColor(QtGui.QPalette.WindowText,
                                CurrentTheme.HOVER_SELECT_COLOR)

    def setCellData(self, cellType, cellId):
        """ setCellData(cellType: str, cellId: int) -> None Create an
        image based on the cell type and id. Then assign it to the
        label. If cellType is None, the cell will be drawn with
        transparent background. If cellType is '', the cell will be
        drawn with the caption 'Empty'. Otherwise, the cell will be
        drawn with white background containing cellType as caption and
        a small rounded shape on the lower right painted with cellId
        
        """
        self.type = cellType
        self.id = cellId
        size = QtCore.QSize(*CurrentTheme.VIRTUAL_CELL_LABEL_SIZE)
        image = QtGui.QImage(size.width() + 12,
                             size.height()+ 12,
                             QtGui.QImage.Format_ARGB32_Premultiplied)
        image.fill(0)

        font = QtGui.QFont()
        font.setStyleStrategy(QtGui.QFont.ForceOutline)
        painter = QtGui.QPainter()
        painter.begin(image)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        if self.type is None:
            painter.setPen(QtCore.Qt.NoPen)
            painter.setBrush(QtCore.Qt.NoBrush)
        else:
            if self.type=='':
                painter.setPen(QtCore.Qt.gray)
                painter.setBrush(QtCore.Qt.NoBrush)
            else:
                painter.setPen(QtCore.Qt.black)
                painter.setBrush(QtCore.Qt.lightGray)
        painter.drawRoundRect(QtCore.QRectF(0.5, 0.5, image.width()-1,
                                            image.height()-1), 25, 25)

        painter.setFont(font)
        if self.type is not None:
            painter.drawText(QtCore.QRect(QtCore.QPoint(6, 6), size),
                             QtCore.Qt.AlignCenter | QtCore.Qt.TextWrapAnywhere,
                             split_camel_case(self.type))
            # Draw the lower right corner number if there is an id
            if self.id>=0 and self.type:
                QAnnotatedPipelineView.drawId(painter, image.rect(), self.id,
                                              QtCore.Qt.AlignRight |
                                              QtCore.Qt.AlignBottom)

        painter.end()

        self.setPixmap(QtGui.QPixmap.fromImage(image))

    def mousePressEvent(self, event):
        """ mousePressEvent(event: QMouseEvent) -> None
        Start the drag and drop when the user click on the label
        
        """
        if self.type:
            mimeData = QtCore.QMimeData()
            mimeData.cellData = (self.type, self.id)

            drag = QtGui.QDrag(self)
            drag.setMimeData(mimeData)
            drag.setHotSpot(self.pixmap().rect().center())
            drag.setPixmap(self.pixmap())
            
            self.setCellData('', -1)
            
            drag.start(QtCore.Qt.MoveAction)
            if mimeData.cellData!=('', -1):
                self.setCellData(*mimeData.cellData)
            self.emit(QtCore.SIGNAL('finishedDragAndDrop'))

    def dragEnterEvent(self, event):
        """ dragEnterEvent(event: QDragEnterEvent) -> None
        Set to accept drops from the other cell info
        
        """
        mimeData = event.mimeData()
        if hasattr(mimeData, 'cellData'):
            event.setDropAction(QtCore.Qt.MoveAction)
            event.accept()
            self.highlight()
        else:
            event.ignore()

    def dropEvent(self, event):        
        """ dropEvent(event: QDragMoveEvent) -> None
        Accept drop event to set the current cell
        
        """
        mimeData = event.mimeData()
        if hasattr(mimeData, 'cellData'):
            event.setDropAction(QtCore.Qt.MoveAction)
            event.accept()            
            if (self.type,self.id)!=(mimeData.cellData[0],mimeData.cellData[1]):
                oldCellData = (self.type, self.id)
                self.setCellData(*mimeData.cellData)
                mimeData.cellData = oldCellData
        else:
            event.ignore()
        self.highlight(False)

    def dragLeaveEvent(self, event):
        """ dragLeaveEvent(event: QDragLeaveEvent) -> None
        Un highlight the current cell
        
        """
        self.highlight(False)

    def highlight(self, on=True):
        """ highlight(on: bool) -> None
        Highlight the cell as if being selected
        
        """
        if on:
            self.setFrameStyle(QtGui.QFrame.Panel | QtGui.QFrame.Plain)
        else:
            self.setFrameStyle(QtGui.QFrame.Panel)

###############################################################################

import unittest

class TestCamelCase(unittest.TestCase):
    def test_simple(self):
        self.assertEqual(split_camel_case('camelCaseH'), 'camel\nCase\nH')
        self.assertEqual(split_camel_case('CamelCase'), 'Camel\nCase')

    def test_long(self):
        self.assertEqual(split_camel_case('parseHTML'), 'parse\nHTML')
        self.assertEqual(split_camel_case('HTMLParser'), 'HTML\nParser')
        self.assertEqual(split_camel_case('vHTMLParser'), 'v\nHTML\nParser')

if __name__=="__main__":
    import sys
    import vistrails.gui.theme
    app = QtGui.QApplication(sys.argv)
    vistrails.gui.theme.initializeCurrentTheme()
    vc = QVirtualCellConfiguration()
    vc.configVirtualCells(['VTKCell', 'ImageViewerCell', 'RichTextCell'])
    vc.show()
    sys.exit(app.exec_())



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Finds all .py files recursively in current directory (.) and updates 2007
year with 2008 in the file header.
"""

from itertools import izip
import os
import re


# The new copyright: the found copyright line will be replaced by this
NEW_COPYRIGHT = ["## Copyright (C) 2014-2015, New York University.\n"]

# The old copyright line: the first matching line in a file will be replaced
# by NEW_COPYRIGHT
RE_COPYRIGHT = re.compile(r"\s*## Copyright \(C\) 20\d\d-20\d\d, "
                          r"New York University\.\s*")

# Number of lines in which to search for the old copyright
NB_SEARCHED_LINES = 5

# List of file names to ignore
IGNORE_LIST = ["update_copyright_year.py"]

files = []
for (path, dnames, fnames) in os.walk('.'):
    for fn in fnames:
        if fn not in IGNORE_LIST and fn.endswith(".py"):
            files.append(os.path.join(path, fn))

# Go through files and update them
print "%d files found" % len(files)
count = 0
for fname in files:
    fp = open(fname, 'rb')
    try:
        # Search only in the first few lines
        for line_num, line in izip(xrange(NB_SEARCHED_LINES), fp):
            if RE_COPYRIGHT.search(line):
                print "Updating %s" % fname
                fp.seek(0)
                lines = fp.readlines()
                fp.close()
                lines[line_num:line_num + 1] = NEW_COPYRIGHT
                fp = file(fname, 'wb')
                fp.writelines(lines)
                count += 1
                break
    finally:
        fp.close()
print "%d files updated" % count



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.vistrail.action import Action
from vistrails.core.log.log import Log
from vistrails.core.vistrail.operation import AddOp, ChangeOp, DeleteOp
from vistrails.db.services.io import SaveBundle
import vistrails.db.services.io
import vistrails.db.services.vistrail
import vistrails.db.services.action
from xml.dom.minidom import parse, getDOMImplementation

def get_db_vistrail_list(config):
    return vistrails.db.services.io.get_db_object_list(config,'vistrail')

def get_workflow(vt, version):
    from vistrails.core.vistrail.pipeline import Pipeline
    workflow = vistrails.db.services.vistrail.materializeWorkflow(vt, version)
    Pipeline.convert(workflow)
    return workflow

def open_workflow(filename):
    from vistrails.core.vistrail.pipeline import Pipeline
    workflow = vistrails.db.services.io.open_workflow_from_xml(filename)
    Pipeline.convert(workflow)
    return workflow

def save_workflow(workflow, filename):
    vistrails.db.services.io.save_workflow_to_xml(workflow, filename)

def save_vistrail_to_xml(vistrail, filename):
    vistrails.db.services.io.save_vistrail_to_xml(vistrail, filename)

def load_vistrail(locator, is_abstraction=False):
    from vistrails.core.vistrail.vistrail import Vistrail

    abstraction_files = []
    thumbnail_files = []
    mashups = []
    vistrail = None
    if locator is None:
        vistrail = Vistrail()
    else:
        res = locator.load()
        if type(res) == type(SaveBundle(None)):
            vistrail = res.vistrail
            abstraction_files.extend(res.abstractions)
            thumbnail_files.extend(res.thumbnails)
            mashups.extend(res.mashups)
        else:
            vistrail = res
    vistrail.is_abstraction = is_abstraction
    return (vistrail, abstraction_files, thumbnail_files, mashups)
    
def open_registry(filename):
    from vistrails.core.modules.module_registry import ModuleRegistry
    registry = vistrails.db.services.io.open_registry_from_xml(filename)
    ModuleRegistry.convert(registry)
    return registry

def unserialize(str, klass):
    """returns VisTrails entity given an XML serialization

    """
    obj = vistrails.db.services.io.unserialize(str, klass.vtType)
    if obj:
        #maybe we should also put a try except here
        klass.convert(obj)
    return obj

def serialize(object):
    """returns XML serialization for any VisTrails entity

    """
    return vistrails.db.services.io.serialize(object)

def open_log(fname, was_appended=False):
    log = vistrails.db.services.io.open_log_from_xml(fname, was_appended)
    Log.convert(log)
    return log


def merge_logs(new_log, log_fname):
    log = vistrails.db.services.io.merge_logs(new_log, log_fname)
    Log.convert(log)
    return log

def get_workflow_diff(vt_pair_1, vt_pair_2):
    """get_workflow_diff( tuple(Vistrail, id), tuple(Vistrail, id) ) ->
            Pipeline, Pipeline, [tuple(id, id)], [tuple(id, id)], 
            [id], [id], [tuple(id, id, list)]

    Return a difference between two workflows referenced as vistrails.
    """

    from vistrails.core.vistrail.pipeline import Pipeline
    (v1, v2, pairs, heuristic_pairs, v1_only, v2_only, param_changes,
     cparam_changes, annot_changes, _, _, _, _) = \
         vistrails.db.services.vistrail.getWorkflowDiff(vt_pair_1, vt_pair_2,
                                                        True)
    Pipeline.convert(v1)
    Pipeline.convert(v2)
    return (v1, v2, pairs, heuristic_pairs, v1_only, v2_only, param_changes,
            cparam_changes, annot_changes)

def get_workflow_diff_with_connections(vt_pair_1, vt_pair_2):
    """get_workflow_diff_with_connections

    Similar to get_workflow_diff but with connection pairings.
    """

    from vistrails.core.vistrail.pipeline import Pipeline
    (v1, v2, m_pairs, m_heuristic, v1_only, v2_only, param_changes,
     cparam_changes, annot_changes, c_pairs, c_heuristic, c1_only, c2_only) =\
         vistrails.db.services.vistrail.getWorkflowDiff(vt_pair_1, vt_pair_2,
                                                        False)
    Pipeline.convert(v1)
    Pipeline.convert(v2)
    return (v1, v2, m_pairs, m_heuristic, v1_only, v2_only, param_changes,
            cparam_changes, annot_changes, c_pairs, c_heuristic, c1_only,
            c2_only)

def getPathAsAction(vt, v1, v2, do_copy=False):
    a = vistrails.db.services.vistrail.getPathAsAction(vt, v1, v2, do_copy)
    Action.convert(a)
    return a

def fixActions(vt, v, actions):
    return vistrails.db.services.vistrail.fixActions(vt, v, actions)

def convert_operation_list(ops):
    for op in ops:
        if op.vtType == 'add':
            AddOp.convert(op)
        elif op.vtType == 'change':
            ChangeOp.convert(op)
        elif op.vtType == 'delete':
            DeleteOp.convert(op)
        else:
            raise TypeError("Unknown operation type '%s'" % op.vtType)

def create_action(action_list):
    """create_action(action_list: list) -> Action
    where action_list is a list of tuples
     (
      type, 
      object, 
      parent_type=None,
      parent_id=None,
      new_obj=None
     )
    and the method returns a *single* action that accomplishes all 
    of the operations.

    Examples: create_action([('add', module1), ('delete', connection2)]
              create_action([('add', param1, 'function', function1),
                             ('change', old_func, new_func, 'module', m1)])
    Note that create_action([('add', module)]) adds a module and *all* of its
    children.
    """
    action = vistrails.db.services.action.create_action(action_list)
    Action.convert(action)
    return action
    
def create_add_op_chain(object, parent=(None, None)):
    """create_add_op_chain(object: object, 
                           parent=(type : str, id : long)) -> [op]
    where [op] is a list of operations to add the given object and its
    children to a workflow.
    """
    ops = vistrails.db.services.action.create_add_op_chain(object, parent)
    convert_operation_list(ops)
    return ops

def create_change_op_chain(old_obj, new_obj, parent=(None,None)):
    """create_change_op_chain(old_obj: object, new_obj: object, 
                              parent=(type : str, id : long)) -> [op]
    where [op] is a list of operations to change the given object and its
    children to the new object in a workflow.
    """
    ops = vistrails.db.services.action.create_change_op_chain(old_obj, new_obj, parent)
    convert_operation_list(ops)
    return ops

def create_delete_op_chain(object, parent=(None, None)):
    """create_delete_op_chain(object: object, 
                              parent=(type : str, id : long)) -> [op]
    where [op] is a list of operations to delete the given object and its
    children from a workflow.
    """
    ops = vistrails.db.services.action.create_delete_op_chain(object, parent)
    convert_operation_list(ops)
    return ops

def create_temp_folder(prefix='vt_save'):
    return vistrails.db.services.io.create_temp_folder(prefix=prefix)

def remove_temp_folder(temp_dir):
    vistrails.db.services.io.remove_temp_folder(temp_dir)

def load_startup(startup_fname):
    from vistrails.core.startup import VistrailsStartup
    startup = vistrails.db.services.io.open_startup_from_xml(startup_fname)
    VistrailsStartup.convert(startup)
    return startup

def save_startup(startup, fname):
    startup = vistrails.db.services.io.save_startup_to_xml(startup, fname)
    return startup
    



#!/usr/bin/env python

# This example demonstrates how to use the vtkPlaneWidget to probe a
# dataset and then generate contours on the probed data.

import vtk

# Create a mace out of filters.
sphere = vtk.vtkSphereSource()
cone = vtk.vtkConeSource()
glyph = vtk.vtkGlyph3D()
glyph.SetInputConnection(sphere.GetOutputPort())
glyph.SetSource(cone.GetOutput())
glyph.SetVectorModeToUseNormal()
glyph.SetScaleModeToScaleByVector()
glyph.SetScaleFactor(0.25)

# The sphere and spikes are appended into a single polydata.
# This just makes things simpler to manage.
apd = vtk.vtkAppendPolyData()
apd.AddInput(glyph.GetOutput())
apd.AddInput(sphere.GetOutput())

maceMapper = vtk.vtkPolyDataMapper()
maceMapper.SetInputConnection(apd.GetOutputPort())

maceActor = vtk.vtkLODActor()
maceActor.SetMapper(maceMapper)
maceActor.VisibilityOn()

# This portion of the code clips the mace with the vtkPlanes
# implicit function. The clipped region is colored green.
plane = vtk.vtkPlane()
clipper = vtk.vtkClipPolyData()
clipper.SetInputConnection(apd.GetOutputPort())
clipper.SetClipFunction(plane)
clipper.InsideOutOn()

selectMapper = vtk.vtkPolyDataMapper()
selectMapper.SetInputConnection(clipper.GetOutputPort())

selectActor = vtk.vtkLODActor()
selectActor.SetMapper(selectMapper)
selectActor.GetProperty().SetColor(0, 1, 0)
selectActor.VisibilityOff()
selectActor.SetScale(1.01, 1.01, 1.01)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# The callback function
def myCallback(obj, event):
    global plane, selectActor
    obj.GetPlane(plane)
    selectActor.VisibilityOn()
 
# Associate the line widget with the interactor
planeWidget = vtk.vtkImplicitPlaneWidget()
planeWidget.SetInteractor(iren)
planeWidget.SetPlaceFactor(1.25)
planeWidget.SetInput(glyph.GetOutput())
planeWidget.PlaceWidget()
planeWidget.AddObserver("InteractionEvent", myCallback)

ren.AddActor(maceActor)
ren.AddActor(selectActor)

# Add the actors to the renderer, set the background and size
ren.SetBackground(1, 1, 1)
renWin.SetSize(300, 300)
ren.SetBackground(0.1, 0.2, 0.4)

# Start interaction.
iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.modules.vistrails_module import Module, InvalidOutput
import copy

##############################################################################
## Order Operator

class ExecuteInOrder(Module):
    """
    The Order Module alows the user to control which sink of a pair of
    sinks ought to executed first.  Connect the "self" port of each
    sink to the corresponding port.  Note that if you have more than
    two sinks, you can string them together by using a string of Order
    modules.
    """

    def update_upstream(self):
        # don't do update until compute!
        pass

    def compute(self):
        # do update_upstream as compute, but sort by key
        for _, connectorList in sorted(self.inputPorts.iteritems()):
            for connector in connectorList:
                connector.obj.update()
        for iport, connectorList in copy.copy(self.inputPorts.items()):
            for connector in connectorList:
                if connector.obj.get_output(connector.port) is InvalidOutput:
                    self.remove_input_connector(iport, connector)


###############################################################################

import unittest

from vistrails.tests.utils import capture_stdout, execute


class TestOrder(unittest.TestCase):
    def test_1(self):
        with capture_stdout() as output:
            self.assertFalse(execute([
                    ('StandardOutput', 'org.vistrails.vistrails.basic', [
                        ('value', [('String', 'one')]),
                    ]),
                    ('StandardOutput', 'org.vistrails.vistrails.basic', [
                        ('value', [('String', 'two')]),
                    ]),
                    ('ExecuteInOrder', 'org.vistrails.vistrails.control_flow', []),
                ],
                [
                    (0, 'self', 2, 'module1'),
                    (1, 'self', 2, 'module2'),
                ]))
        self.assertEqual(output, ['one', 'two'])

    def test_2(self):
        with capture_stdout() as output:
            self.assertFalse(execute([
                    ('StandardOutput', 'org.vistrails.vistrails.basic', [
                        ('value', [('String', 'two')]),
                    ]),
                    ('StandardOutput', 'org.vistrails.vistrails.basic', [
                        ('value', [('String', 'one')]),
                    ]),
                    ('ExecuteInOrder', 'org.vistrails.vistrails.control_flow', []),
                ],
                [
                    (1, 'self', 2, 'module1'),
                    (0, 'self', 2, 'module2'),
                ]))
        self.assertEqual(output, ['one', 'two'])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from vistrails.core.modules.vistrails_module import Module
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.modules.basic_modules import List, String

from engine_manager import EngineManager
from map import Map


def initialize(*args,**keywords):
    reg = get_module_registry()

    reg.add_module(Map)
    reg.add_input_port(Map, 'FunctionPort', (Module, ''))
    reg.add_input_port(Map, 'InputList', (List, ''))
    reg.add_input_port(Map, 'InputPort', (List, ''))
    reg.add_input_port(Map, 'OutputPort', (String, ''))
    reg.add_output_port(Map, 'Result', (List, ''))


def finalize():
    EngineManager.cleanup()


def menu_items():
    return (
            ("Start new engine processes",
             lambda: EngineManager.start_engines()),
            ("Show information on the cluster",
             lambda: EngineManager.info()),
            ("Change profile",
             lambda: EngineManager.change_profile()),
            ("Cleanup started processes",
             lambda: EngineManager.cleanup()),
            ("Request cluster shutdown",
             lambda: EngineManager.shutdown_cluster()),
    )



#!/usr/bin/env python

# This example demonstrates the use of the vtkTransformPolyDataFilter
# to reposition a 3D text string.

import vtk
from vtk.util.colors import *

# Define a Single Cube
Scalars = vtk.vtkFloatArray()
Scalars.InsertNextValue(1.0)
Scalars.InsertNextValue(0.0)
Scalars.InsertNextValue(0.0)
Scalars.InsertNextValue(1.0)
Scalars.InsertNextValue(0.0)
Scalars.InsertNextValue(0.0)
Scalars.InsertNextValue(0.0)
Scalars.InsertNextValue(0.0)

Points = vtk.vtkPoints()
Points.InsertNextPoint(0, 0, 0)
Points.InsertNextPoint(1, 0, 0)
Points.InsertNextPoint(1, 1, 0)
Points.InsertNextPoint(0, 1, 0)
Points.InsertNextPoint(0, 0, 1)
Points.InsertNextPoint(1, 0, 1)
Points.InsertNextPoint(1, 1, 1)
Points.InsertNextPoint(0, 1, 1)

Ids = vtk.vtkIdList()
Ids.InsertNextId(0)
Ids.InsertNextId(1)
Ids.InsertNextId(2)
Ids.InsertNextId(3)
Ids.InsertNextId(4)
Ids.InsertNextId(5)
Ids.InsertNextId(6)
Ids.InsertNextId(7)

Grid = vtk.vtkUnstructuredGrid()
Grid.Allocate(10, 10)
Grid.InsertNextCell(12, Ids)
Grid.SetPoints(Points)
Grid.GetPointData().SetScalars(Scalars)

# Find the triangles that lie along the 0.5 contour in this cube.
Marching = vtk.vtkContourFilter()
Marching.SetInput(Grid)
Marching.SetValue(0, 0.5)
Marching.Update()

# Extract the edges of the triangles just found.
triangleEdges = vtk.vtkExtractEdges()
triangleEdges.SetInputConnection(Marching.GetOutputPort())
# Draw the edges as tubes instead of lines.  Also create the associated
# mapper and actor to display the tubes.
triangleEdgeTubes = vtk.vtkTubeFilter()
triangleEdgeTubes.SetInputConnection(triangleEdges.GetOutputPort())
triangleEdgeTubes.SetRadius(.005)
triangleEdgeTubes.SetNumberOfSides(6)
triangleEdgeTubes.UseDefaultNormalOn()
triangleEdgeTubes.SetDefaultNormal(.577, .577, .577)
triangleEdgeMapper = vtk.vtkPolyDataMapper()
triangleEdgeMapper.SetInputConnection(triangleEdgeTubes.GetOutputPort())
triangleEdgeMapper.ScalarVisibilityOff()
triangleEdgeActor = vtk.vtkActor()
triangleEdgeActor.SetMapper(triangleEdgeMapper)
triangleEdgeActor.GetProperty().SetDiffuseColor(lamp_black)
triangleEdgeActor.GetProperty().SetSpecular(.4)
triangleEdgeActor.GetProperty().SetSpecularPower(10)

# Shrink the triangles we found earlier.  Create the associated mapper
# and actor.  Set the opacity of the shrunken triangles.
aShrinker = vtk.vtkShrinkPolyData()
aShrinker.SetShrinkFactor(1)
aShrinker.SetInputConnection(Marching.GetOutputPort())
aMapper = vtk.vtkPolyDataMapper()
aMapper.ScalarVisibilityOff()
aMapper.SetInputConnection(aShrinker.GetOutputPort())
Triangles = vtk.vtkActor()
Triangles.SetMapper(aMapper)
Triangles.GetProperty().SetDiffuseColor(banana)
Triangles.GetProperty().SetOpacity(.6)

# Draw a cube the same size and at the same position as the one
# created previously.  Extract the edges because we only want to see
# the outline of the cube.  Pass the edges through a vtkTubeFilter so
# they are displayed as tubes rather than lines.
CubeModel = vtk.vtkCubeSource()
CubeModel.SetCenter(.5, .5, .5)
Edges = vtk.vtkExtractEdges()
Edges.SetInputConnection(CubeModel.GetOutputPort())
Tubes = vtk.vtkTubeFilter()
Tubes.SetInputConnection(Edges.GetOutputPort())
Tubes.SetRadius(.01)
Tubes.SetNumberOfSides(6)
Tubes.UseDefaultNormalOn()
Tubes.SetDefaultNormal(.577, .577, .577)
# Create the mapper and actor to display the cube edges.
TubeMapper = vtk.vtkPolyDataMapper()
TubeMapper.SetInputConnection(Tubes.GetOutputPort())
CubeEdges = vtk.vtkActor()
CubeEdges.SetMapper(TubeMapper)
CubeEdges.GetProperty().SetDiffuseColor(khaki)
CubeEdges.GetProperty().SetSpecular(.4)
CubeEdges.GetProperty().SetSpecularPower(10)

# Create a sphere to use as a glyph source for vtkGlyph3D.
Sphere = vtk.vtkSphereSource()
Sphere.SetRadius(0.04)
Sphere.SetPhiResolution(20)
Sphere.SetThetaResolution(20)
# Remove the part of the cube with data values below 0.5.
ThresholdIn = vtk.vtkThresholdPoints()
ThresholdIn.SetInput(Grid)
ThresholdIn.ThresholdByUpper(.5)
# Display spheres at the vertices remaining in the cube data set after
# it was passed through vtkThresholdPoints.
Vertices = vtk.vtkGlyph3D()
Vertices.SetInputConnection(ThresholdIn.GetOutputPort())
Vertices.SetSource(Sphere.GetOutput())
# Create a mapper and actor to display the glyphs.
SphereMapper = vtk.vtkPolyDataMapper()
SphereMapper.SetInputConnection(Vertices.GetOutputPort())
SphereMapper.ScalarVisibilityOff()
CubeVertices = vtk.vtkActor()
CubeVertices.SetMapper(SphereMapper)
CubeVertices.GetProperty().SetDiffuseColor(tomato)
CubeVertices.GetProperty().SetDiffuseColor(tomato)

# Define the text for the label
caseLabel = vtk.vtkVectorText()
caseLabel.SetText("Case 1")

# Set up a transform to move the label to a new position.
aLabelTransform = vtk.vtkTransform()
aLabelTransform.Identity()
aLabelTransform.Translate(-0.2, 0, 1.25)
aLabelTransform.Scale(.05, .05, .05)

# Move the label to a new position.
labelTransform = vtk.vtkTransformPolyDataFilter()
labelTransform.SetTransform(aLabelTransform)
labelTransform.SetInputConnection(caseLabel.GetOutputPort())
  
# Create a mapper and actor to display the text.
labelMapper = vtk.vtkPolyDataMapper()
labelMapper.SetInputConnection(labelTransform.GetOutputPort())
 
labelActor = vtk.vtkActor()
labelActor.SetMapper(labelMapper)
 
# Define the base that the cube sits on.  Create its associated mapper
# and actor.  Set the position of the actor.
baseModel = vtk.vtkCubeSource()
baseModel.SetXLength(1.5)
baseModel.SetYLength(.01)
baseModel.SetZLength(1.5)
baseMapper = vtk.vtkPolyDataMapper()
baseMapper.SetInputConnection(baseModel.GetOutputPort())
base = vtk.vtkActor()
base.SetMapper(baseMapper)
base.SetPosition(.5, -0.09, .5)

# Create the Renderer, RenderWindow, and RenderWindowInteractor
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
renWin.SetSize(640, 480)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer
ren.AddActor(triangleEdgeActor)
ren.AddActor(base)
ren.AddActor(labelActor)
ren.AddActor(CubeEdges)
ren.AddActor(CubeVertices)
ren.AddActor(Triangles)

# Set the background color.
ren.SetBackground(slate_grey)

# This sets up the right values for case12 of the marching cubes
# algorithm (routine translated from vtktesting/mccases.tcl).
def case12(scalars, caselabel, IN, OUT):
    scalars.InsertValue(0, OUT)
    scalars.InsertValue(1, IN)
    scalars.InsertValue(2, OUT)
    scalars.InsertValue(3, IN)
    scalars.InsertValue(4, IN)
    scalars.InsertValue(5, IN)
    scalars.InsertValue(6, OUT)
    scalars.InsertValue(7, OUT)
    if IN == 1:
        caselabel.SetText("Case 12 - 00111010")
    else:
        caselabel.SetText("Case 12 - 11000101")

# Set the scalar values for this case of marching cubes.
case12(Scalars, caseLabel, 0, 1)

# Force the grid to update.
Grid.Modified()

# Position the camera.
ren.ResetCamera()
ren.GetActiveCamera().Dolly(1.2)
ren.GetActiveCamera().Azimuth(30)
ren.GetActiveCamera().Elevation(20)
ren.ResetCameraClippingRange()

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import vistrails.core.db.action
from vistrails.core.db.locator import XMLFileLocator
from vistrails.core.db.io import serialize, unserialize
from vistrails.core import debug
from vistrails.core.interpreter.default import get_default_interpreter
from vistrails.core.log.group_exec import GroupExec
from vistrails.core.log.machine import Machine
from vistrails.core.log.module_exec import ModuleExec
from vistrails.core.modules.basic_modules import Constant
import vistrails.core.modules.module_registry
import vistrails.core.modules.utils
from vistrails.core.modules.vistrails_module import Module, ModuleError, \
    InvalidOutput
from vistrails.core.vistrail.annotation import Annotation
from vistrails.core.vistrail.controller import VistrailController
from vistrails.core.vistrail.group import Group
from vistrails.core.vistrail.module_function import ModuleFunction
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.vistrail.vistrail import Vistrail
from vistrails.db.domain import IdScope
import vistrails.db.versions

import copy
import inspect
from itertools import izip
import os
import re
import sys
import tempfile

from IPython.parallel.error import CompositeError

from .api import get_client

try:
    import hashlib
    sha1_hash = hashlib.sha1
except ImportError:
    import sha
    sha1_hash = sha.new


###############################################################################
# This function is sent to the engines which execute it
#
# It receives the workflow, and the list of targeted output ports
#
# It returns the corresponding computed outputs and the execution log
#
def execute_wf(wf, output_port):
    # Save the workflow in a temporary file
    temp_wf_fd, temp_wf = tempfile.mkstemp()

    try:
        f = open(temp_wf, 'w')
        f.write(wf)
        f.close()
        os.close(temp_wf_fd)

        # Clean the cache
        interpreter = get_default_interpreter()
        interpreter.flush()

        # Load the Pipeline from the temporary file
        vistrail = Vistrail()
        locator = XMLFileLocator(temp_wf)
        workflow = locator.load(Pipeline)

        # Build a Vistrail from this single Pipeline
        action_list = []
        for module in workflow.module_list:
            action_list.append(('add', module))
        for connection in workflow.connection_list:
            action_list.append(('add', connection))
        action = vistrails.core.db.action.create_action(action_list)

        vistrail.add_action(action, 0L)
        vistrail.update_id_scope()
        tag = 'parallel flow'
        vistrail.addTag(tag, action.id)

        # Build a controller and execute
        controller = VistrailController()
        controller.set_vistrail(vistrail, None)
        controller.change_selected_version(vistrail.get_version_number(tag))
        execution = controller.execute_current_workflow(
                custom_aliases=None,
                custom_params=None,
                extra_info=None,
                reason='API Pipeline Execution')

        # Build a list of errors
        errors = []
        pipeline = vistrail.getPipeline(tag)
        execution_errors = execution[0][0].errors
        if execution_errors:
            for key in execution_errors:
                module = pipeline.modules[key]
                msg = '%s: %s' %(module.name, execution_errors[key])
                errors.append(msg)

        # Get the execution log from the controller
        try:
            module_log = controller.log.workflow_execs[0].item_execs[0]
        except IndexError:
            errors.append("Module log not found")
            return dict(errors=errors)
        else:
            machine = controller.log.workflow_execs[0].machines[
                    module_log.machine_id]
            xml_log = serialize(module_log)
            machine_log = serialize(machine)

        # Get the output value
        output = None
        if not execution_errors:
            executed_module, = execution[0][0].executed
            executed_module = execution[0][0].objects[executed_module]
            try:
                output = executed_module.get_output(output_port)
            except ModuleError:
                errors.append("Output port not found: %s" % output_port)
                return dict(errors=errors)
            if isinstance(output, Module):
                raise TypeError("Output value is a Module instance")

        # Return the dictionary, that will be sent back to the client
        return dict(errors=errors,
                    output=output,
                    xml_log=xml_log,
                    machine_log=machine_log)
    finally:
        os.unlink(temp_wf)

###############################################################################

_ansi_code = re.compile(r'%s(?:(?:\[[^A-Za-z]*[A-Za-z])|[^\[])' % '\x1B')

def strip_ansi_codes(s):
    return _ansi_code.sub('', s)

###############################################################################
# Map Operator
#
class Map(Module):
    """The Map Module executes a map operator in parallel on IPython engines.

    The FunctionPort should be connected to the 'self' output of the module you
    want to execute.
    The InputList is the list of values to be scattered on the engines.
    """
    def __init__(self):
        Module.__init__(self)

    def update_upstream(self):
        """A modified version of the update_upstream method."""

        # everything is the same except that we don't update anything
        # upstream of FunctionPort
        for port_name, connector_list in self.inputPorts.iteritems():
            if port_name == 'FunctionPort':
                for connector in connector_list:
                    connector.obj.update_upstream()
            else:
                for connector in connector_list:
                    connector.obj.update()
        for port_name, connectorList in copy.copy(self.inputPorts.items()):
            if port_name != 'FunctionPort':
                for connector in connectorList:
                    if connector.obj.get_output(connector.port) is \
                            InvalidOutput:
                        self.remove_input_connector(port_name, connector)

    @staticmethod
    def print_compositeerror(e):
        sys.stderr.write("Got %d exceptions from IPython engines:\n" %
                         len(e.elist))
        for e_type, e_msg, formatted_tb, infos in e.elist:
            sys.stderr.write("Error from engine %d (%r):\n" % (
                             infos['engine_id'], infos['engine_uuid']))
            sys.stderr.write("%s\n" % strip_ansi_codes(formatted_tb))

    @staticmethod
    def list_exceptions(e):
        return '\n'.join(
                "% 3d: %s: %s" % (infos['engine_id'],
                                  e_type,
                                  e_msg)
                for e_type, e_msg, tb, infos in e.elist)

    def updateFunctionPort(self):
        """
        Function to be used inside the updateUsptream method of the Map module. It
        updates the module connected to the FunctionPort port, executing it in
        parallel.
        """
        nameInput = self.get_input('InputPort')
        nameOutput = self.get_input('OutputPort')
        rawInputList = self.get_input('InputList')

        # Create inputList to always have iterable elements
        # to simplify code
        if len(nameInput) == 1:
            element_is_iter = False
            inputList = [[element] for element in rawInputList]
        else:
            element_is_iter = True
            inputList = rawInputList

        workflows = []
        module = None
        vtType = None

        # iterating through the connectors
        for connector in self.inputPorts.get('FunctionPort'):
            module = connector.obj

            # pipeline
            original_pipeline = connector.obj.moduleInfo['pipeline']

            # module
            module_id = connector.obj.moduleInfo['moduleId']
            vtType = original_pipeline.modules[module_id].vtType

            # serialize the module for each value in the list
            for i, element in enumerate(inputList):
                if element_is_iter:
                    self.element = element
                else:
                    self.element = element[0]

                # checking type and setting input in the module
                self.typeChecking(connector.obj, nameInput, inputList)
                self.setInputValues(connector.obj, nameInput, element, i)

                pipeline_db_module = original_pipeline.modules[module_id].do_copy()

                # transforming a subworkflow in a group
                # TODO: should we also transform inner subworkflows?
                if pipeline_db_module.is_abstraction():
                    group = Group(id=pipeline_db_module.id,
                                  cache=pipeline_db_module.cache,
                                  location=pipeline_db_module.location,
                                  functions=pipeline_db_module.functions,
                                  annotations=pipeline_db_module.annotations)

                    source_port_specs = pipeline_db_module.sourcePorts()
                    dest_port_specs = pipeline_db_module.destinationPorts()
                    for source_port_spec in source_port_specs:
                        group.add_port_spec(source_port_spec)
                    for dest_port_spec in dest_port_specs:
                        group.add_port_spec(dest_port_spec)

                    group.pipeline = pipeline_db_module.pipeline
                    pipeline_db_module = group

                # getting highest id between functions to guarantee unique ids
                # TODO: can get current IdScope here?
                if pipeline_db_module.functions:
                    high_id = max(function.db_id
                                  for function in pipeline_db_module.functions)
                else:
                    high_id = 0

                # adding function and parameter to module in pipeline
                # TODO: 'pos' should not be always 0 here
                id_scope = IdScope(beginId=long(high_id+1))
                for elementValue, inputPort in izip(element, nameInput):

                    p_spec = pipeline_db_module.get_port_spec(inputPort, 'input')
                    descrs = p_spec.descriptors()
                    if len(descrs) != 1:
                        raise ModuleError(
                                self,
                                "Tuple input ports are not supported")
                    if not issubclass(descrs[0].module, Constant):
                        raise ModuleError(
                                self,
                                "Module inputs should be Constant types")
                    type = p_spec.sigstring[1:-1]

                    mod_function = ModuleFunction(id=id_scope.getNewId(ModuleFunction.vtType),
                                                  pos=0,
                                                  name=inputPort)
                    mod_param = ModuleParam(id=0L,
                                            pos=0,
                                            type=type,
                                            val=elementValue)

                    mod_function.add_parameter(mod_param)
                    pipeline_db_module.add_function(mod_function)

                # serializing module
                wf = self.serialize_module(pipeline_db_module)
                workflows.append(wf)

            # getting first connector, ignoring the rest
            break

        # IPython stuff
        try:
            rc = get_client()
        except Exception, error:
            raise ModuleError(self, "Exception while loading IPython: %s" %
                              debug.format_exception(error))
        if rc is None:
            raise ModuleError(self, "Couldn't get an IPython connection")
        engines = rc.ids
        if not engines:
            raise ModuleError(
                    self,
                    "Exception while loading IPython: No IPython engines "
                    "detected!")

        # initializes each engine
        # importing modules and initializing the VisTrails application
        # in the engines *only* in the first execution on this engine
        uninitialized = []
        for eng in engines:
            try:
                rc[eng]['init']
            except Exception:
                uninitialized.append(eng)
        if uninitialized:
            init_view = rc[uninitialized]
            with init_view.sync_imports():
                import tempfile
                import inspect

                # VisTrails API
                import vistrails
                import vistrails.core
                import vistrails.core.db.action
                import vistrails.core.application
                import vistrails.core.modules.module_registry
                from vistrails.core.db.io import serialize
                from vistrails.core.vistrail.vistrail import Vistrail
                from vistrails.core.vistrail.pipeline import Pipeline
                from vistrails.core.db.locator import XMLFileLocator
                from vistrails.core.vistrail.controller import VistrailController
                from vistrails.core.interpreter.default import get_default_interpreter

            # initializing a VisTrails application
            try:
                init_view.execute(
                        'app = vistrails.core.application.init('
                        '        {"spawned": True},'
                        '        args=[])',
                        block=True)
            except CompositeError, e:
                self.print_compositeerror(e)
                raise ModuleError(self, "Error initializing application on "
                                  "IPython engines:\n"
                                  "%s" % self.list_exceptions(e))

            init_view['init'] = True

        # setting computing color
        module.logging.set_computing(module)

        # executing function in engines
        # each map returns a dictionary
        try:
            ldview = rc.load_balanced_view()
            map_result = ldview.map_sync(execute_wf, workflows, [nameOutput]*len(workflows))
        except CompositeError, e:
            self.print_compositeerror(e)
            raise ModuleError(self, "Error from IPython engines:\n"
                              "%s" % self.list_exceptions(e))

        # verifying errors
        errors = []
        for engine in range(len(map_result)):
            if map_result[engine]['errors']:
                msg = "ModuleError in engine %d: '%s'" % (
                        engine,
                        ', '.join(map_result[engine]['errors']))
                errors.append(msg)

        if errors:
            raise ModuleError(self, '\n'.join(errors))

        # setting success color
        module.logging.signalSuccess(module)

        reg = vistrails.core.modules.module_registry.get_module_registry()
        self.result = []
        for map_execution in map_result:
            output = map_execution['output']
            self.result.append(output)

        # including execution logs
        for engine in range(len(map_result)):
            log = map_result[engine]['xml_log']
            exec_ = None
            if (vtType == 'abstraction') or (vtType == 'group'):
                exec_ = unserialize(log, GroupExec)
            elif (vtType == 'module'):
                exec_ = unserialize(log, ModuleExec)
            else:
                # something is wrong...
                continue

            # assigning new ids to existing annotations
            exec_annotations = exec_.annotations
            for i in range(len(exec_annotations)):
                exec_annotations[i].id = self.logging.log.log.id_scope.getNewId(Annotation.vtType)

            parallel_annotation = Annotation(key='parallel_execution', value=True)
            parallel_annotation.id = self.logging.log.log.id_scope.getNewId(Annotation.vtType)
            annotations = [parallel_annotation] + exec_annotations
            exec_.annotations = annotations

            # before adding the execution log, we need to get the machine information
            machine = unserialize(map_result[engine]['machine_log'], Machine)
            machine_id = self.logging.add_machine(machine)

            # recursively add machine information to execution items
            def add_machine_recursive(exec_):
                for item in exec_.item_execs:
                    if hasattr(item, 'machine_id'):
                        item.machine_id = machine_id
                        if item.vtType in ('abstraction', 'group'):
                            add_machine_recursive(item)

            exec_.machine_id = machine_id
            if (vtType == 'abstraction') or (vtType == 'group'):
                add_machine_recursive(exec_)

            self.logging.add_exec(exec_)


    def serialize_module(self, module):
        """
        Serializes a module to be executed in parallel.
        """

        def process_group(group):
            group.pipeline.id = None
            for module in group.pipeline.module_list:
                if module.is_group():
                    process_group(module)

        pipeline = Pipeline(version=vistrails.db.versions.currentVersion)

        if module.is_group():
            process_group(module)

        module = module.do_copy()
        pipeline.add_module(module)

        return serialize(pipeline)

    def compute(self):
        """The compute method for Map."""

        self.result = None
        self.updateFunctionPort()

        self.set_output('Result', self.result)

###############################################################################

class NewConstant(Constant):
    """
    A new Constant module to be used inside the Map module.
    """
    def setValue(self, v):
        self.set_output("value", v)
        self.upToDate = True

def create_constant(value):
    """
    Creates a NewConstant module, to be used for the ModuleConnector.
    """
    constant = NewConstant()
    constant.setValue(value)
    return constant

def get_module(value, signature):
    """
    Creates a module for value, in order to do the type checking.
    """

    from vistrails.core.modules.basic_modules import Boolean, String, Integer, Float, List

    if isinstance(value, Constant):
        return type(value)
    elif isinstance(value, bool):
        return Boolean
    elif isinstance(value, str):
        return String
    elif isinstance(value, int):
        return Integer
    elif isinstance(value, float):
        return Float
    elif isinstance(value, list):
        return List
    elif isinstance(value, tuple):
        v_modules = ()
        for element in xrange(len(value)):
            v_modules += (get_module(value[element], signature[element]))
        return v_modules
    else:
        from vistrails.core import debug
        debug.warning("Could not identify the type of the list element.")
        debug.warning("Type checking is not going to be done inside Map module.")
        return None



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import math

################################################################################
# Point

class Point(object):
    """Point is a point in 2D. It behaves as a vector, too, because that's
    convenient."""

    ##########################################################################
    # Constructor
    
    def __init__(self, x=0, y=0):
        """ Point(x: float, y: float) -> Point
        Initialize and return a Point
        
        """
        self.x = float(x)
        self.y = float(y)

    # copy is implicit - all fields are immutable

    ##########################################################################

    def reset(self,x,y):
        """ reset(x: float, y: float) -> None
        Resets the point to given coordinates and return nothing

        """
        self.x = float(x)
        self.y = float(y)

    def is_inside(self, rect):
        """ is_inside(rect: Rect) -> boolean
        Check if the point is falling inside rect and return a boolean
        
        """
        return (self.x >= rect.lower_left.x and
                self.x <= rect.upper_right.x and
                self.y >= rect.lower_left.y and
                self.y <= rect.upper_right.y)
    
    def length(self):
        """ length() -> float
        Interprets self as a vector to compute the L_2 norm and return a float

        """
        return math.sqrt(self.x * self.x + self.y * self.y)

    ##########################################################################
    # Debugging

    def show_comparison(self, other):
        if type(self) != type(other):
            print "Type mismatch"
            return
        l = (self - other).length()
        if l >= self.eq_delta:
            print "Points are too far away:"
            print self
            print other
        else:
            print "No difference found: delta is %f" % l
            assert self == other

    ##########################################################################
    # Operators

    def __str__(self):
        return "(%f, %f)@%X" % (self.x, self.y, id(self))

    def __neg__(self):
        """ __neg__() -> Point
        Compute a point p such that: self + p == Point(0,0), and return a Point
        
        """
        return Point(-self.x,-self.y)

    def __add__(self, other):
        """ __add__(other: Point) -> Point
        Returns a point p such that: self + other == p, and return a Point
        
        """
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        """ __sub__(other: Point) -> Point
        Returns a point p such that: self - other == p, and return a Point

        """
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        """ __mul__(other: float) -> Point
        Interprets self as a vector to perform a scalar multiplication and
        return a Point

        """
        return Point(self.x * other, self.y * other)

    eq_delta = 0.0001
    def __eq__(self, other):
        """__eq__(other: Point) -> boolean 
        Two points are equal if they have the same components 
        
        """
        if type(self) != type(other):
            return False
        return (self.x - other.x) ** 2 + (self.y - other.y) ** 2 < 1e-8
        
    def __ne__(self, other):
        """__ne__(other: Point) -> boolean 
        Two points are differenr if they don't have the same components 
        
        """
        return not self.__eq__(other)

    def __rmul__(self, other):
        """ __rmul__(other: float) -> Point
        Interprets self as a vector to perform a scalar multiplication and
        return a Point

        """
        return Point(self.x * other, self.y * other)


################################################################################
# Unit tests

import unittest
import random

class TestPoint(unittest.TestCase):

    @staticmethod
    def assert_double_equals(a, b, eps = 0.00001):
        assert abs(a-b) < eps

    def test_add_length(self):
        """Uses triangle inequality to exercise add and length"""
        for i in xrange(100):
            x = Point(random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0))
            y = Point(random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0))
            assert (x+y).length() <= x.length() + y.length()

    def test_mul_length(self):
        """Uses vector space properties to exercise mul, rmul and length"""
        for i in xrange(100):
            x = Point(random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0))
            s = random.uniform(0.0, 10.0)
            self.assert_double_equals(s * x.length(), (s * x).length())
            self.assert_double_equals(s * x.length(), (x * s).length())

    def test_comparison_operators(self):
        """ Test comparison operators """
        a = Point(0, 1)
        b = Point(0, 1)
        assert a == b
        assert a is not None
        b = Point(0, 0.1)
        assert a != b

if __name__ == '__main__':
    unittest.main()
        



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from collections import namedtuple as _namedtuple, Mapping
from itertools import izip

_docs = {}

def get_field_name(v):
    if isinstance(v[0], tuple):
        return v[0][0]
    return v[0]

def namedtuple(typename, fields):
    field_names = [get_field_name(f) for f in fields]
    default_values = []
    default_found = False
    field_types = []
    docstrings = []

    new_fields = []
    for f in fields:
        field_name = get_field_name(f)
        field_key = (typename, field_name)
        field_type = None
        docstring = None
        default_value = None
        if field_key in _docs:
            (field_type, docstring) = _docs[field_key]
        else:
            if len(f) > 1 and f[1]:
                field_type = f[1]
            if len(f) > 2 and f[2]:
                docstring = f[2]

        if isinstance(f[0], tuple):
            default_found = True
            default_value = f[0][1]
            default_values.append(default_value)
        elif default_found:
            raise Exception('Field "%s" must have default value since the '
                            'preceding field does.' % f[0])
        if field_type:
            field_types.append("(%s)" % field_type)
        else:
            field_types.append("")
        if docstring:
            docstrings.append("\n        %s\n" % docstring)
        else:
            docstrings.append("")

        new_fields.append(((field_name, default_value), field_type, docstring))
    
    fields = new_fields

    T = _namedtuple(typename, field_names)
    args = []
    if len(default_values) > 0:
        T.__new__.__defaults__ = tuple(default_values)
        args.extend("%s=%s" % (field_name, '"%s"' % default_val \
                               if isinstance(default_val, basestring) \
                               else default_val) \
                    for (field_name, default_val) in \
                    izip(reversed(field_names), reversed(default_values)))
        args.reverse()
        args = field_names[:len(field_names)-len(default_values)] + args
    else:
        args = field_names
    
    init_docstring = ""
    for (field_name, field_type, docstring) in \
        izip(field_names, field_types, docstrings):
        init_docstring += "    .. py:attribute:: %s %s\n%s\n" % \
                          (field_name, field_type, docstring)

    init_template = "def __init__(self, %s): pass" % ', '.join(args)
    d = {}
    exec init_template in d
    T.__init__ = d['__init__']
    T.__init__.im_func.__doc__ = init_docstring
    T._vistrails_fields = fields

    return T
    

def subnamedtuple(typename, cls, new_fields=[], override_fields=[]):
    override_dict = dict((get_field_name(f), f) for f in override_fields)
    fields = []
    for f in cls._vistrails_fields:
        field_name = get_field_name(f)
        if field_name in override_dict:
            fields.append(override_dict[field_name])
        else:
            fields.append(f)
    fields.extend(new_fields)

    return namedtuple(typename, fields)
    
_documentation = \
"""
   ConstantWidgetConfig.widget: QWidget | String

      The widget to be used either as the class itself or a string of
      the form ``<py_module>:<class>``
      (e.g. "vistrails.gui.modules.constant_configuration:BooleanWidget")

   ConstantWidgetConfig.widget_type: String

      A developer-available type that links a widget to a port entry_type

   ConstantWidgetConfig.widget_use: String

      Intended to differentiate widgets for different purposes in
      VisTrails.  Currently uses the values None, "query", and
      "paramexp" to define widget uses in parameter, query, and
      parameter exploration configurations, repsectively.

   QueryWidgetConfig.widget_use: String

      Like :py:attr:`ConstantWidgetConfig.widget_use`, just defaulted
      to "query"

   ParamExpWidgetConfig.widget_use: String

      Like :py:attr:`ConstantWidgetConfig.widget_use`, just defaulted
      to "paramexp"

   ModuleSettings.name: String

      An optional string that will identify the module (defaults to
      the module class's ``__name__``).

   ModuleSettings.configure_widget: QWidget | PathString

      An optional module configuration widget that is provided so that
      users can configure the module (e.g. PythonSource uses a widget
      to display a code editor more suited for writing code).

   ModuleSettings.constant_widget: ConstantWidgetConfig | QWidget | PathString

      If not None, the registry will use the specified widget(s) or
      import path string(s) of the form "<module>:<class>" to import
      the widget.  A tuple allows a user to specify the widget_type
      (e.g. "default", "enum", etc.)  as the second item and the
      widget_use (e.g. "default", "query", "paramexp").  If the plural
      form is used, we expect a list, the singular form is used for a
      single item.

   ModuleSettings.constant_widgets: List(constant_widget)

      See :py:attr:`ModuleSettings.constant_widget`

   ModuleSettings.signature: Callable

      If not None, then the cache uses this callable as the function
      to generate the signature for the module in the cache. The
      function should take three parameters: the pipeline (of type
      vistrails.core.vistrail.Pipeline), the module (of type
      vistrails.core.vistrail.Module), and a dict that stores
      parameter hashers. This dict is supposed to be passed to
      vistrails/core/cache/hasher.py:Hasher, in case that needs to be
      called.

   ModuleSettings.constant_signature: Callable

      If not None, the cache uses this callable as the function used
      to generate the signature for parameters of the associated
      module's type.  This function should take a single argument (of
      type vistrials.core.vistrail.module_param.ModuleParam) and
      returns a SHA hash.

   ModuleSettings.color: (Float, Float, Float)

      In the GUI, the module will be colored according to the
      specified RGB tuple where each value is a float in [0,1.0].

   ModuleSettings.fringe: List((Float, Float))

      If not None, generates custom lateral fringes for the module
      boxes.  module_fringe must be a list of pairs of floats.  The
      first point must be (0.0, 0.0), and the last must be (0.0,
      1.0). This will be used to generate custom lateral fringes for
      module boxes. All x values must be positive, and all y values
      must be between 0.0 and 1.0. Alternatively, the user can set
      :py:attr:`ModuleSettings.left_fringe` and
      :py:attr:`ModuleSettings.right_fringe` to set two different
      fringes.

   ModuleSettings.left_fringe: List((Float, Float))

      See :py:attr:`ModuleSettings.fringe`.

   ModuleSettings.right_fringe: List((Float, Float))

      See :py:attr:`ModuleSettings.fringe`.

   ModuleSettings.abstract: Boolean

      If True, means the module only serves as a base class for other
      modules.  It will not appear in the module palette but may be
      used as a input or output port type.

   ModuleSettings.package: String

      If not None, then we use this package instead of the current
      one.  This is only intended to be used with local per-module
      module registries (in other words: if you don't know what a
      local per-module registry is, you can ignore this option).

   ModuleSettings.namespace: String

      If not None, then we associate a namespace with the module. A
      namespace is essentially appended to the package identifier so
      that multiple modules inside the same package can share the same
      name.  Namespaces may be nested using the "|" separator.  For
      example, "Tools|Mathematical" specified the Mathematical
      namespace inside of the Tools namespace.

   ModuleSettings.version: String

      The module version.  This is usually used with subworkflows
      where the underlying module may have different versions.  Not
      recommended for general use at this time.

   ModuleSettings.package_version: String

      The current package version for the module.  As with
      :py:attr:`ModuleSettings.package`, you shuold not use this
      unless you know what you are doing as VisTrails will
      automatically fill in this information for any normal package.

   ModuleSettings.hide_namespace: Boolean

      If True, the module palette will not display the namespace for
      the module.  Used for subworkflows, otherwise it may be
      confusing for users.

   ModuleSettings.hide_descriptor: Boolean

      If True, the module palette will not display that module in its
      list (similar to abstract, but can be used when the module is
      not truly abstract).

   ModuleSettings.is_root: Boolean

      Internal use only.  This is used to designate the base Module
      class and should not be used by any other module.

   ModuleSettings.ghost_package: String

      If not None, then the 'ghost_identifier' is set on the
      descriptor, which will cause the module to be displayed under
      that package in the module palette, rather than the package
      specified by the package argument (or current package).

   ModuleSettings.ghost_package_version: String

      If not None, then the attribute 'ghost_package_version' is set
      on the descriptor.  Currently this value is unused, but
      eventually if multiple packages with the same identifier but
      different package versions are loaded simultaneously, this will
      allow overriding the package_version to clean up the module
      palette.

   ModuleSettings.ghost_namespace: String

      If not None, the descriptor will be displayed under the
      specified namespace instead of the 'namespace' attribute of the
      descriptor.

   Port.name: String

      The name of the of the port

   Port.signature: String

      The signature of the of the port (e.g. "basic:Float")

   Port.optional: Boolean

      Whether the port should be visible by default (defaults to True)

   Port.sort_key: Integer

      An integer value that indicates where, relative to other ports,
      this port should appear visually

   Port.docstring: String

      Documentation for the port

   Port.shape: "triangle" | "diamond" | "circle" | [(Float,Float)]

      The shape of the port.  If triangle, appending an angle (in
      degrees) rotates the triangle.  If a list of (x,y) tuples,
      specifies points of a polygon in the [0,1] x [0,1] region

   Port.min_conns: Integer

      The minimum number of values required for the port

   Port.max_conns: Integer

      The maximum number of values allowed for the port

   Port.depth: Integer

      The list depth of the port. Default is 0 (no list)
      
   InputPort.label: String

      A label to be shown with a port

   InputPort.default:

      The default value for a constant-typed port

   InputPort.values: List

      A list of enumerated values that a
      :py:class:`.ConstantWidgetConfig` uses to configure the widget.
      For example, the "enum" widget uses the entries to present an
      exclusive list of choices.

   InputPort.entry_type: String

      The type of the configuration widget that should be used with
      this port. Developers may use custom widgets on a port-by-port
      basis by adding widgets with different
      :py:attr:`ConstantWidgetConfig.widget_type` values and defining
      the port\'s entry_type to match.

   CompoundInputPort.labels: List(String)

      A list of :py:attr:`InputPort.label`

   CompoundInputPort.defaults: List

      A list of :py:attr:`InputPort.default`

   CompoundInputPort.values: List(List)

      A list of :py:attr:`InputPort.values`

   CompoundInputPort.entry_types: List(String)

      A list of :py:attr:`InputPort.entry_type`

   CompoundInputPort.items: List(InputPortItem)

      Either use this field and break individual
      labels/defaults/values/entry_types into
      :py:class:`.InputPortItem` components or use the other four
      fields.

   CompoundInputPort.signature: String

      The signature of the port (e.g. "basic:Integer, basic:Float").
      Note that for compound ports, this may be instead included on a
      per-component basis in :py:attr:`InputPortItem.signature`.

   CompoundOutputPort.items: List(OutputPortItem)

      Either use this field and break individual signatures into
      :py:class:`.OutputPortItem` components or use the signature
      field.

   CompoundOutputPort.signature: String

      The signature of the port (e.g. "basic:Integer, basic:Float").
      Note that for compound ports, this may be instead included on a
      per-component basis in :py:attr:`OutputPortItem.signature`.

   PortItem.signature: String

      See :py:attr:`InputPort.signature`

   InputPortItem.label: String

      See :py:attr:`InputPort.label`

   InputPortItem.default:

      See :py:attr:`InputPort.default`

   InputPortItem.values: List

      See :py:attr:`InputPort.values`

   InputPortItem.entry_type: String

      See :py:attr:`InputPort.entry_type`

"""

def parse_documentation():
    line_iter = iter(_documentation.splitlines())
    line_iter.next()
    for line in line_iter:
        field, field_type = line.strip().split(':', 1)
        (cls_name, field_name) = field.split('.')
        doc_lines = []
        line = line_iter.next()
        while True:
            line = line_iter.next()
            if not line.strip():
                break
            doc_lines.append(line.strip())
        _docs[(cls_name, field_name)] = (field_type, ' '.join(doc_lines))

parse_documentation()

ConstantWidgetConfig = \
    namedtuple('ConstantWidgetConfig', 
               [('widget',),
                (('widget_type', None),),
                (('widget_use', None),)])
QueryWidgetConfig = subnamedtuple('QueryWidgetConfig', ConstantWidgetConfig, 
    override_fields=[(('widget_use', 'query'),)])
ParamExpWidgetConfig = subnamedtuple('ParamExpWidgetConfig', 
    ConstantWidgetConfig,
    override_fields=[(('widget_use', 'paramexp'),)])

ModuleSettings = namedtuple('ModuleSettings',
                          [(('name', None),),
                           (('configure_widget', None),),
                           (('constant_widget', None),),
                           (('constant_widgets', None),),
                           (('signature', None),),
                           (('constant_signature', None),),
                           (('color', None),),
                           (('fringe', None),),
                           (('left_fringe', None),),
                           (('right_fringe', None),),
                           (('abstract', False),),
                           (('package', None),),
                           (('namespace', None),),
                           (('version', None),),
                           (('package_version', None),),
                           (('hide_namespace', False),),
                           (('hide_descriptor', False),),
                           (('is_root', False),),
                           (('ghost_package', None),),
                           (('ghost_package_version', None),),
                           (('ghost_namespace', None),),])

Port = namedtuple('Port', 
                     [("name",),
                      ("signature",),
                      (("optional", False),),
                      (("sort_key", -1),),
                      (("docstring", None),),
                      (("shape", None),),
                      (("min_conns", 0),),
                      (("max_conns", -1),),
                      (("depth", 0),),
                      ])

InputPort = subnamedtuple('InputPort', Port,
                          [(('label', None),),
                           (('default', None),),
                           (('values', None),),
                           (('entry_type', None),)])
OutputPort = subnamedtuple('OutputPort', Port)

CompoundPort = subnamedtuple('CompoundPort', Port,
                             [(('items', None),)],
                             [(('signature', None),)])
CompoundInputPort = subnamedtuple('CompoundInputPort', CompoundPort,
                                  [(('labels', None),), 
                                   (('defaults', None),),
                                   (('values', None),),
                                   (('entry_types', None),)])
CompoundOutputPort = subnamedtuple('CompoundOutputPort', CompoundPort)

PortItem = namedtuple('PortItem', [('signature', None)])
InputPortItem = subnamedtuple('InputPortItem', PortItem,
                              [(('label', None),),
                               (('default', None),),
                               (('values', None),),
                               (('entry_type', None),)])
OutputPortItem = subnamedtuple('OutputPortItem', PortItem)

DeprecatedInputPort = \
                      namedtuple('DeprecatedInputPort',
                                 [("name", "String",),
                                  ("signature", "String",),
                                  (("optional", True),),
                                  (("sort_key", -1),),
                                  (('labels', None),), 
                                  (('defaults', None),),
                                  (('values', None),),
                                  (('entry_types', None),),
                                  (("docstring", None),),
                                  (("shape", None),),
                                  (("min_conns", 0),),
                                  (("max_conns", -1),),
                              ])

IPort = InputPort
OPort = OutputPort
CIPort = CompoundInputPort
COPort = CompoundOutputPort
IPItem = InputPortItem
OPItem = OutputPortItem



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This code creates layouts for workflow digraphs.
Originally written by Lauro D. Lins.

"""

####################################################

from __future__ import division

def uniquify(seq, idfun=None): 
    # order preserving
    if idfun is None:
        def idfun(x): return x
    seen = {}
    result = []
    for item in seq:
        marker = idfun(item)
        # in old Python versions:
        # if seen.has_key(marker)
        # but in new ones:
        if marker in seen: continue
        seen[marker] = 1
        result.append(item)
    return result


class Defaults(object):
    u            = 10.0
    label_margin = 20.0 

class Vec2(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y
    def set(self, x, y):
        self.x = x
        self.y = y
    def __add__(self, other):
        if isinstance(other, (float, int)):
            return Vec2(self.x + other, self.y + other)
        else:
            return Vec2(self.x + other.x, self.y + other.y)
    def __radd__(self, other):
        return self.__add__(other)
    def __sub__(self, other):
        if isinstance(other, (float, int)):
            return Vec2(self.x - other, self.y - other)
        else:
            return Vec2(self.x - other.x, self.y - other.y)
    def __rsub__(self, other):
        return self.__sub__(other)
    def __mul__(self, other):
        if isinstance(other, (float, int)):
            return Vec2(self.x * other, self.y * other)
        else:
            return Vec2(self.x * other.x, self.y * other.y)
    def __rmul__(self, other):
        return self.__mul__(other)
    def dot(self, other):
        return self.x * other.x + self.y * other.y
    def __str__(self):
        return "(%.2f, %.2f)" % (self.x, self.y)
    def __repr__(self):
        return "(%.2f, %.2f)" % (self.x, self.y)

####################################################

UNDEFINED_LAYER = None
INPUT_PORT, OUTPUT_PORT = 1,2

class Pipeline(object):

    def __init__(self):
        self.modules     = []
        self.connections = []

    def createModule(self, shortname, name, num_input_ports, num_output_ports, prev_x=None):
        mod = Module(self, len(self.modules), shortname, name, num_input_ports, num_output_ports, prev_x)
        self.modules.append(mod)
        return mod

    def createConnection(self, mod_source, port_source_idx, mod_target, port_target_idx):     
        source_port = mod_source.getOutputPort(port_source_idx)
        target_port = mod_target.getInputPort(port_target_idx)

        conn = Connection(self, source_port, target_port)
        source_port.addConnection(conn)
        target_port.addConnection(conn)

        self.connections.append(conn)

        return conn

    # def updateCache(self):
    #     self.cached_num_modules      = len(self.modules)
    #     self.cached_num_connections  = len(self.connections)



class Module(object):

    def __init__(self, workflow, key, shortname, name, num_input_ports, num_output_ports, prev_x=None):

        self.workflow     = workflow
        self.key          = key
        self.shortname    = shortname
        self.name         = name

        self.input_ports  = [ Port(self, INPUT_PORT,  i) for i in xrange(num_input_ports ) ]
        self.output_ports = [ Port(self, OUTPUT_PORT, i) for i in xrange(num_output_ports) ]

        self.layout_layer_number = UNDEFINED_LAYER
        self.layout_layer_index  = 0

        self.cached              = False
        self.cached_in_degree    = 0
        self.cached_out_degree   = 0

        self.cached_connections_succ  = [] # successor modules
        self.cached_connections_pred  = [] # predecessor modules

        self.cached_succ         = [] # successor modules
        self.cached_pred         = [] # predecessor modules

        self.cached_num_succ     = 0
        self.cached_num_pred     = 0

        self.flags               = 0

        self.layout_pos          = Vec2(0,0)
        self.layout_dim          = Vec2(1,1)
        
        self.prev_x = prev_x

    def clearFlags(self):
        self.flags = 0

    def updateCache(self):
        self.cached_in_degree = self.getInDegree()
        self.cached_out_degree = self.getOutDegree()

        self.cached_connections_succ  = [] # successor modules
        self.cached_connections_pred  = [] # predecessor modules

        # cached succ
        self.cached_succ = []
        for p in self.output_ports:
            p.collectOppositeModules(self.cached_succ)
            self.cached_connections_succ.extend([c for c in p.connections])

        uniquify(self.cached_succ)
        # print "%s's of cached_succ: " % (self.shortname),
        # print [m.name for m in self.cached_succ]


        # cached pred
        self.cached_pred = []
        for p in self.input_ports:
            p.collectOppositeModules(self.cached_pred)
            self.cached_connections_pred.extend([c for c in p.connections])
        uniquify(self.cached_pred)

        self.cached_num_succ     = len(self.cached_succ)
        self.cached_num_pred     = len(self.cached_pred)

        self.cached = True

    def getInputPort(self, idx):
        return self.input_ports[idx]
                              
    def getOutputPort(self, idx):
        return self.output_ports[idx]

    def getInDegree(self):
        count = 0
        for ip in self.input_ports:
            count += ip.getNumConnections()
        return count

    def getOutDegree(self):
        count = 0
        for ip in self.output_ports:
            count += ip.getNumConnections()
        return count



class Connection(object):
    def __init__(self, workflow, source_port, target_port):
        self.workflow    = workflow
        self.source_port = source_port # an output port
        self.target_port = target_port # an input port

    def getOppositePort(self, port):
        if self.source_port == port:
            return self.target_port
        elif self.target_port == port:
            return self.source_port
        else:
            raise ValueError("oppositePort problem: no opposite port")



class Port(object):

    def __init__(self, module, kind, index):
        self.module      = module
        self.kind        = kind
        self.index       = index
        self.connections = []

        self.layout_pos  = Vec2(0,0)
        self.layout_dim  = Vec2(1,1)

    def getNumConnections(self):
        return len(self.connections)

    def addConnection(self, connection):
        self.connections.append(connection)

        # module is not cached
        self.module.cached = False

    def collectOppositeModules(self, lst):
        for c in self.connections:
            lst.append(c.getOppositePort(self).module)



class ItemDFS(object):
    def __init__(self, module, index, max_index):
        self.module    = module
        self.index     = index
        self.max_index = max_index

        # the creation of ItemDFS implies
        # that it will be stacked on the DFSIterator
        # and the module flag goes from 0 to 1
        self.module.flags = 1



class DFSIterator(object):

    def __init__(self, workflow, update_module_cache=False):
        self.stack = []

        # stack source modules
        for m in workflow.modules:

            m.clearFlags()

            if not m.cached:
                m.updateCache()

            if m.cached_in_degree == 0: # source modules
                m.flags = 1
                self.stack.append(ItemDFS(m,-1,m.cached_num_succ))


    def next(self):
        
        while len(self.stack) > 0:

            top = self.stack[-1]
            top.index += 1
            mod = top.module
            
            #DEBUG
            #print "pop %s index %d max_index %d" % (mod.shortname, top.index, top.max_index)

            if top.index >= top.max_index:
                self.stack.pop()
                return mod
            else:
                next_mod = mod.cached_succ[top.index]
                if next_mod.flags > 0:
                    continue
                next_mod.flags = 1
                self.stack.append(ItemDFS(next_mod, -1, next_mod.cached_num_succ))

        return None

class Layer(object):

    def __init__(self, index):
        self.index   = index
        self.modules = []

        self.y = 0.0

        self.max_module_height = 0.0
        self.sum_modules_width = 0.0

    def addModule(self, module):
        self.modules.append(module)

        self.sum_modules_width += module.layout_dim.x
        self.max_module_height  = max(self.max_module_height,module.layout_dim.y)

        # print "addModule(self, module): max_module_height = ", self.max_module_height



class Layers(object):

    def __init__(self):
        self.layers  = []

    def addModule(self, module, layer_number):

        num_layers = len(self.layers)

        # print "num_layers",    num_layers
        # print "layer_number ", layer_number
        # print "layers",self.layers
        
        if layer_number >= num_layers:
            self.layers.extend([Layer(i) for i in xrange(num_layers,layer_number+1)])
        
        layer = self.layers[layer_number]
        layer.addModule(module)

    
class Page(object):
    def __init__(self, page_x0, page_y0, page_width, page_height):
        self.x0     = page_x0
        self.y0     = page_y0
        self.width  = page_width
        self.height = page_height

    def __str__(self):
        # print self.x0
        # print self.y0
        # print self.width
        # print self.height
        return "%.2f %.2f %.2f %.2f" % (self.x0, self.y0, self.width, self.height)

class WorkflowLayout(object):
    def __init__(self, wf, module_size_f, module_margin, port_size,
                 port_interspace):
        self.wf = wf
        self.module_size_f = module_size_f
        self.module_margin = module_margin
        self.port_size = port_size
        self.port_interspace = port_interspace

    #
    # Compute sizes
    # 
    def compute_module_sizes(self): 
        wf = self.wf # wf is a workflow

        u = Defaults.u # u is the base unit in points to derive the other sizes 

        #
        # A module visual representation consists of three rows
        #
        # first row: input ports
        # mid   row: text label
        # third row: output ports
        #
        row_margin = 0.20 * u
        row_height = u + 2 * row_margin

        module_height       =  3 * row_height # three stripes of

        #DEBUG
        #print "module_height = ", module_height

        module_text_margin  = u
        module_ports_margin = row_margin

        # port_margin         = 0.30 * u
        # port_interspace     = 0.15 * u
        # port_size           = 0.75 * u

        for module in wf.modules:
            module_width, module_height = self.module_size_f(module)

            # [DK] why do we have to do this here?
            if not module.cached:
                module.updateCache()

            max_num_ports = max(len(module.input_ports), len(module.output_ports))

            # ports_width = 2*module_ports_margin + max_num_ports * u + (max_num_ports-1) * module_ports_margin 
            # text_width  = 2*module_text_margin  + len(module.name) * u * 0.6

            # module_width = max(ports_width, text_width)

            module.layout_pos.set(0,0)
            module.layout_dim.set(module_width, module_height)

            for port_list, do_reverse, d in [(module.input_ports, False, -1), 
                                             (module.output_ports, True, 1)]:
                if do_reverse:
                    port_iter = reversed(port_list)
                else:
                    port_iter = port_list

                for i, p in enumerate(port_iter):
                    p.layout_pos.set(d * module_width/2.0 -
                                     d * self.module_margin[0] -
                                     d * (i + 0.5) * self.port_size[0] -
                                     d * self.port_interspace * i,
                                     d * (module_height/2.0 -
                                          self.module_margin[1] - 
                                          self.port_size[1]))
                    p.layout_dim.set(*self.port_size)
                                     
            # for p in module.input_ports:
            #     p.layout_pos.set(-module_width/2.0    +  # initial position
            #                       port_margin         +  # ports margin
            #                       (p.index + 0.5) * port_size +  # widths of previous ports + half port width
            #                       port_interspace * p.index,     # sum of previous interspace
            #                       -module_height / 2.0 + row_height / 2.0 )
            #     p.layout_dim.set(*port_size)


            # op = list(module.output_ports)
            # op.reverse()
            # for i in xrange(len(op)):
            #     p = op[i]
            #     p.layout_pos.set( module_width/2.0 - 
            #                       port_margin - 
            #                       (i + 0.5) * port_size -
            #                       port_interspace * i,
            #                       +module_height/2.0 - row_height / 2.0 )
            #     p.layout_dim.set(port_size, port_size)

    def assign_modules_to_layers(self):
        wf = self.wf

        iterator = DFSIterator(wf)

        # topologically sorted permulation of the modules
        permutation = []
        while True:
            mod = iterator.next()
            if mod is None:
                break
            permutation.append(mod)
        permutation.reverse()

        # define layers
        i = 0
        for module in permutation:
            if module.cached_num_pred == 0:
                module.layout_layer_number = 0
            else:
                module.layout_layer_number = 1 + max([pred.layout_layer_number for pred in module.cached_pred])

            # use permutation index as a the sort key to define
            # the initial permutation of the modules in each layer
            module.layout_layer_index = i
            i += 1


        # log
        # for module in permutation:
        #     print "layer of %s = %d" % (module.shortname, module.layout_layer_number)

        # adjust free modules (a free module in this context is
        # a module whose minimum successor layer and maximum predecessor layer
        # have more than one intermediate layer in between

        #
        # TODO: adjust layer
        for module in permutation:
            if module.cached_num_pred == 0 and module.cached_num_succ > 0:
                module.layout_layer_number = min([succ.layout_layer_number for succ in module.cached_succ]) - 1
            # else:
            #     a = min([succ.layout_layer_number for succ in module.cached_succ]) - 1
            # print "layer of %s = %d" % (module.shortname, module.layout_layer_number)

    def assign_module_to_layers_no_gaps(self):
        if len(self.wf.modules) == 0:
            return
        
        visited = set()
        min_layer = [0]
        def set_module_layer_number(module, layer_number):
            module.layout_layer_number = layer_number
            visited.add(module)
            min_layer[0] = min(min_layer[0], layer_number)
            for port in module.input_ports:
                for conn in port.connections:
                    if conn.source_port.module not in visited:
                        set_module_layer_number(conn.source_port.module, layer_number-1)
            for port in module.output_ports:
                for conn in port.connections:
                    if conn.target_port.module not in visited:
                        set_module_layer_number(conn.target_port.module, layer_number+1)
               
        set_module_layer_number(self.wf.modules[0], 0)
                        
        #adjust all layers numbers so that the min is 0
        if min_layer[0] < 0:
            for module in self.wf.modules:
                module.layout_layer_number -= min_layer[0]

    def assign_module_permutation_to_each_layer(self, preserve_order=False):
        wf = self.wf

        # create layers
        layers = Layers()
        for module in wf.modules:
            layers.addModule(module, module.layout_layer_number)

        # sort modules by the current value of layout_layer_index
        for layer in layers.layers:
            layer.modules.sort(lambda a,b: a.layout_layer_index - b.layout_layer_index)
            for i in xrange(len(layer.modules)):
                layer.modules[i].layout_layer_index = i

        #
        num_layers = len(layers.layers)

        lastModified = [-1 for i in xrange(num_layers)]

        #
        # sweep down and up reducing the number of crossings
        # using the barycentric method (heuristic)
        #
        iteration = 0
        updates = len(wf.modules)
        while updates > 0 and iteration < 100:

            # print "iteration: ", iteration

            updates = 0

            DOWN, UP = 1, -1
            for direction in [DOWN, UP]:

                # print "   direction:", direction

                delta = 1 if direction == DOWN else -1

                i0, i1 = (0,num_layers) if direction == DOWN else (num_layers-1,-1)

                # sweep "direction"
                for i in xrange(i0+delta,i1,direction):
                # for i in xrange(2,4):


                    i_prev = i-delta
                    # print "      layer:"        , i
                    # print "         layer_prev:", i_prev

                    layer = layers.layers[i]

                    # if one module on layer then continue
                    num_modules = len(layer.modules)
                    if num_modules == 1:
                        continue

                    # if nothing chaged on this layer and the "previous" layer (in the sweep direction)
                    # in the last iteration then nothing needs to be done
                    if iteration - lastModified[i_prev] > 1 and \
                           iteration - lastModified[i] > 1:
                        continue

                    # apply barycentric permutation to layer "i" using neighbors on layer "i + delta"
                    barycenters = [-1] * num_modules
                    for j in xrange(num_modules):

                        module      = layer.modules[j]
                        connections = \
                            module.cached_connections_pred  \
                            if direction == DOWN else       \
                            module.cached_connections_succ

                        value = 0.0
                        for c in connections:
                            opposite_port = c.source_port if direction == DOWN else c.target_port
                            value += opposite_port.module.layout_layer_index + opposite_port.index/100.0
                        if len(connections) > 0:
                            value /= 1.0 * len(connections)
                            barycenters[j] = value

                    for j in xrange(1,num_modules):
                        if barycenters[j] < 0:
                            barycenters[j] = barycenters[j-1] + 1e-5

                    # print "         barycenters: ", barycenters


                    new_order = [(barycenters[j], layer.modules[j]) for j in xrange(num_modules)]
                    new_order.sort()

                    # print "            indices before: ", [u.layout_layer_index for u in layer.modules]
                    # print "            indices after:  ", [aux[0] for aux in new_order]
                    # print "            names before =  ",[mod.name for mod in layer.modules]

                    for j in xrange(num_modules):
                        module = new_order[j][1]
                        if module.layout_layer_index != j:
                            # print "            module indexed %d goes to index %d" % (module.layout_layer_index,j)
                            lastModified[i] = iteration
                            updates += 1

                    if lastModified[i] == iteration:
                        for j in xrange(num_modules):
                            module = new_order[j][1]
                            module.layout_layer_index = j


                    # sort using the last layout_layer_index
                    layer.modules.sort(lambda a,b: a.layout_layer_index - b.layout_layer_index)
                    # print "            names after =",[mod.name for mod in layer.modules]


                # print "      updates: ", updates

            # iteration
            iteration += 1

            # break
            break
        
        if preserve_order:
            for layer in layers.layers:                

                # sort using the last layout_layer_index
                layer.modules.sort(key=lambda x: x.layout_layer_index)
            
                #separate modules that have no previous x value
                temp = []
                for i in reversed(range(len(layer.modules))):
                    if layer.modules[i].prev_x is None:
                        temp.append((i,layer.modules.pop(i)))
                
                #sort on previous x
                layer.modules.sort(key=lambda x: x.prev_x)
                
                #put separated modules back in their original slot
                for item in reversed(temp):
                    layer.modules.insert(item[0],item[1])
                    
                #reassign index
                for i in range(len(layer.modules)):
                    layer.modules[i].layout_layer_index = i
            
    #
    # this method is "friend" of the classes above in the C++ sense:
    # it can access and modify
    #
    def compute_layout(self, layer_x_separation, layer_y_separation): 
        wf = self.wf # wf is a workflow

        # create layers
        layers = Layers()
        for module in wf.modules:
            layers.addModule(module, module.layout_layer_number)

        for layer in layers.layers:# sort using the last layout_layer_index
            layer.modules.sort(lambda a,b: a.layout_layer_index - b.layout_layer_index)

        # spread layers
        min_x = max_x = 0.0
        y = min_y = max_y = 0.0

        top_layer = True

        for layer in layers.layers:

            # print "layer: ", layer.index

            if not top_layer:
                y += layer_y_separation
            else:
                top_layer = False


            layer_half_height = layer.max_module_height / 2.0

            # print "layer_half_height", layer_half_height

            y += layer_half_height


            # print "layer ", layer.index, " y = ", y

            layer_num_modules = len(layer.modules)
            layer_width = layer.sum_modules_width + (layer_num_modules-1) * layer_x_separation
            layer_min_x = -layer_width/2.0

            x = layer_min_x

            # set module positions
            for i in xrange(layer_num_modules):

                if (i > 0):
                    x += layer_x_separation

                module = layer.modules[i]

                module_half_width = module.layout_dim.x / 2.0

                x += module_half_width

                module.layout_pos.set(x, y)

                x += module_half_width

            layer_max_x = x

            min_x = min(min_x, layer_min_x)
            max_x = max(max_x, layer_max_x)

            y += layer_half_height

        max_y = y # y only got smaller


        #
        page_x_margin = 3 * Defaults.u
        page_y_margin = 3 * Defaults.u

        page_util_width  = max_x - min_x
        page_util_height = max_y - min_y

        page_width  = page_util_width  + page_x_margin * 2
        page_height = page_util_height + page_y_margin * 2

        page_x0     = -page_width /2.0
        page_y0     = -page_y_margin

        page = Page(page_x0, page_y0, page_width, page_height)

        return page

    def run_all(self, layer_x_separation=50, layer_y_separation=50, preserve_order=False, no_gaps=False):
        self.compute_module_sizes()
        if no_gaps:
            self.assign_module_to_layers_no_gaps()
        else:
            self.assign_modules_to_layers()
        self.assign_module_permutation_to_each_layer(preserve_order)
        self.compute_layout(layer_x_separation, layer_y_separation)


#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import os
import shutil

from parse import run as run_parse
from diff import compute_diff, apply_diff
from generate import run as run_generate

_bases = ["artists", "plots"]

def backup():
    if not os.path.exists("backup"):
        os.mkdir("backup")
    backup_files = ["mpl_%s_raw.xml", "mpl_%s.xml", "mpl_%s_diff.xml"]
    for base in _bases:
        for fname_base in backup_files:
            fname = fname_base % base
            shutil.copy(fname, "backup")

def run(which="all"):
    backup()
    if which == "all":
        bases = _bases
    else:
        bases = [which]
    for base in bases:
        compute_diff("mpl_%s_raw.xml" % base, 
                     "mpl_%s.xml" % base, 
                     "mpl_%s_diff.xml" % base)
        run_parse(base)
        apply_diff("mpl_%s_raw.xml" % base,
                   "mpl_%s_diff.xml" % base,
                   "mpl_%s.xml" % base)
        run_generate(base)

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        if sys.argv[1] == 'backup':
            backup()
        else:
            run(sys.argv[1])
    else:
        run()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""
This module handles Parameter Exploration in VisTrails
"""
from __future__ import division

from vistrails.core import debug
from vistrails.core.vistrail.module_function import ModuleFunction
from vistrails.core.vistrail.module_param import ModuleParam
import copy

import unittest

################################################################################

class ParameterExploration(object):
    """
    ParameterExploration is a class that will take a pipeline and a
    set of parameter values and executes all of them

    """
    
    def __init__(self, specs):
        """ ParameterExploration(specs: list) -> ParameterExploration
        Takes a list of interpolator list. The number of items in the
        list is also the number of dimensions this parameter
        exploration is going to explore on
        
        """
        self.specs = specs

    def explore(self, pipeline):
        """ explore(pipeline: VisPipeline) -> list[VisPipeline]        
        Apply parameter exploration on multiple dimensions using the
        values in self.specs
        
        """
        pipelineList = [pipeline]
        for i in xrange(len(self.specs)):
            pipelineList = self.interpolateList(pipelineList, self.specs[i])
        return pipelineList

    def interpolateList(self, pipelineList, interpList):
        """ interpolateList(pipeline: list[VisPipeLine],
                            interpList: InterpolateDiscreteParam)
                            -> list[VisPipeline]                            
        Returns a list of interpolated pipelines, by applying the
        interpolators in a multiplication manner
        
        """
        if len(interpList)<1: return pipelineList
        stepCount = interpList[0].stepCount
        result = []
        for step in xrange(stepCount):
            for pipeline in pipelineList:
                newp = copy.copy(pipeline)
                for interp in interpList:
                    interp.perform(newp, step)
                result.append(newp)
        return result
    
class InterpolateDiscreteParam(object):
    """
    InterpolateDiscreteParam takes in ranges and a number of
    steps. Then given a pipeline, it will interpolate and generate a
    number of steps pipelines with values interpolated between the
    ranges
    
    """

    def __init__(self, module, function, ranges, stepCount):
        """ InterpolateDiscreteParam(module: Module,
                                     function: str,
                                     ranges: list(tuple),
                                     stepCount: int)
                                     -> InterpolateDiscreteParam
                                     
        Initialize the interpolator with a specific module function
        given a list of parameter ranges

        Keyword arguments:
        module    --- module id in a pipeline        
        function  --- a string express the function name that belongs
                      to module
                      
        ranges    --- [tupe(min,max) or tuple(s1,s2..,s{stepCount}],
                      ranges specified for each argument of function
                      where s{i} is of type 'str'
                      
        stepCount --- the number of step for the interpolation
                                     
        """
        self.module = module
        self.function = function
        self.stepCount = stepCount
        self.values = self.interpolate(ranges, stepCount)

    def interpolate(self, ranges, stepCount):
        """ interpolate(ranges: tuple, stepCount: int) -> list
        
        This function takes a number of (min,max) or (s1,...sn) to
        interpolate exact stepCount number of step. The output will be
        a list of stepCount elements where each of them is (a1,...,an).
        a{i} is either int or string and n is the number of arguments.
        
        """
        params = []
        for r in ranges:
            interpolatedValues = []
            argumentType = type(r[0])
            if argumentType in [int, float]:
                for i in xrange(stepCount):
                    if stepCount>1: t = i/float(stepCount-1)
                    else: t = 0
                    interpolatedValues.append(argumentType(r[0]+t*(r[1]-r[0])))
            elif argumentType==str:
                interpolatedValues = list(r)
            else:
                debug.critical('Cannot interpolate non-cardinal types')
                assert False
            params.append(interpolatedValues)
        return zip(*params)
        
        
    def perform(self, pipeline, step):
        """ perform(pipeline: VisPipeline, step: int) -> None        
        This will takes a pipeline and apply the interpolated values
        at step 'step' to the pipeline. Then return the updated
        pipeline

        """
        m = pipeline.modules[self.module.id]
        f = ModuleFunction()
        f.name = self.function
        f.returnType = 'void'
        value = self.values[step]
        for v in value:
            p = ModuleParam()
            convert = {'int':'Integer', 'str':'String',
                       'float':'Float', 'double':'Float'}
            p.type = convert[type(v).__name__]
            p.strValue = str(v)
            f.params.append(p)
        m.functions.append(f)

class ActionBasedParameterExploration(object):
    """
    ActionBasedParameterExploration is going to replace
    ParameterExploration class to perform parameter exploration. It is
    an action based parameter exploration
    
    """
    def explore(self, pipeline, actions, pre_actions=[]):
        """ explore(pipeline: Pipeline, actions: [action set],
                           pre_actions: [action set]) -> [(pipeline, actions)]
        Perform the parameter exploration on the pipeline with a set
        of actions. Each 'action set' represent a tuple of actions that we
        need to apply on the pipeline at a step in a dimension. For example:
        actions = [[(action1a, action2a), (action1b, action2b)],
                   [(action3a), (action3b)]]
                   
        which means this is a 2 x 1 parameter exploration resulting of
        executing 2 pipelines. The first one is
        pipeline.action1a.action2a.action3a and the second one is
        pipeline.action1b.action2b.action3b.
        
        pre_actions are actions that should be applied first to all pipelines
        and usually contains creation of functions that do not already exist

        The return values are a list of tuples containing interpolated
        pipeline and a set of actions leading to that interpolated
        pipeline. This is useful for update the parameter exploration
        back to the builder.
        
        """
        results = []
        resultActions = []
        
        def exploreDimension(pipeline, performedActions, dim):
            """ exploreDimension(pipeline: Pipeline, performedActions: [actions],
                                 dim: int) -> None
            Start applying actions to the pipeline at dimension
            dim. 'pipeline' will not be modified in the function
            
            """
            if dim<0:
                results.append(pipeline)
                resultActions.append(performedActions)
                return
            currentActions = actions[dim]
            if len(currentActions)==0:
                # Ignore empty dimension
                exploreDimension(pipeline, performedActions, dim-1)
                return
            for actionSet in currentActions:
                currentPipeline = copy.copy(pipeline)
                currentPeformedActions = copy.copy(performedActions)
                for action in actionSet:
                    currentPipeline.perform_action(action)
                    currentPeformedActions.append(action)
                exploreDimension(currentPipeline, currentPeformedActions, dim-1)
        
        # perform pre_actions
        currentPipeline = copy.copy(pipeline)
        for action in pre_actions:
            currentPipeline.perform_action(action)
        
        exploreDimension(currentPipeline, pre_actions, len(actions)-1)
        return (results, resultActions)

def _pipelinePositions(sheetCount, rowCount, colCount,
                       pipelines):
    """ _pipelinePositions(sheetCount: int, rowCount: int,
                           colCount: int, pipelines: list of Pipeline,
                           controller: VistrailCintroller
                           -> list of Positions
    Apply the pipeline locations to a list of pipeline positions in a
    parameter exploration given that pipelines has multiple chunk
    of sheetCount x rowCount x colCount cells

    """

    pipelinePositions = []
    for pId in xrange(len(pipelines)):
        col = pId % colCount
        row = (pId // colCount) % rowCount
        sheet = (pId // (colCount*rowCount)) % sheetCount
        pipelinePositions.append((row, col, sheet))
    return pipelinePositions


################################################################################
        

class TestParameterExploration(unittest.TestCase):
    """
    Test if ParameterExploration is executing correctly. For now it is a very
    simple test to test more of the interpolated values
    
    """
    def testInterpolator(self):
        interpolator = InterpolateDiscreteParam(0, 'testing',
                                                [(0,10),
                                                 (0.0,10.0),
                                                 ('one', 'two', 'three')],
                                                3)
        self.assertEqual(interpolator.values,
                         [(0, 0.0, 'one'),
                          (5, 5.0, 'two'),
                          (10, 10.0, 'three')])

if __name__ == '__main__':
    unittest.main()
